__all__ = [
    "cimac_id_to_cimac_participant_id",
    "get_property",
    "identity",
    "MetadataModel",
    "with_default_session",
]

from functools import wraps
from typing import Any, Dict, Optional, Tuple

from flask import current_app
from sqlalchemy.orm import Session

from ...config.db import BaseModel

# some simple functions to handle common process_as cases
# the parameters for all process_as functions are
# value : Any, context : dict[Column, Any]
identity = lambda v, _: v
cimac_id_to_cimac_participant_id = lambda cimac_id, _: cimac_id[:7]

# this is a special-case handler, where any property in context
# can be retrieved by using get_property(< key >) as a process_as function
# # created to get the `object_url` of files after GCS URI formatting
get_property = lambda prop: lambda _, context: (
    [v for k, v in context.items() if k.name == prop] + [None]
)[0]


def with_default_session(f):
    """
    For some `f` expecting a database session instance as a keyword argument,
    set the default value of the session keyword argument to the current app's
    database driver's session. We need to do this in a decorator rather than
    inline in the function definition because the current app is only available
    once the app is running and an application context has been pushed.
    """

    @wraps(f)
    def wrapped(*args, **kwargs):
        if "session" not in kwargs:
            kwargs["session"] = current_app.extensions["sqlalchemy"].db.session
        return f(*args, **kwargs)

    return wrapped


class MetadataModel(BaseModel):
    __abstract__ = True

    def primary_key_values(self) -> Optional[Tuple[Any]]:
        """
        Returns a tuple of the values of the primary key values
        Special value None if all of the pk columns are None
        """
        pk_map = self.primary_key_map()
        return tuple(pk_map.values()) if pk_map is not None else None

    def primary_key_map(self) -> Optional[Dict[str, Any]]:
        """
        Returns a dict of Column: value for any primary key column.
        Special value None if all of the pk columns are None
        """
        from .utils import _all_bases

        columns_to_check = [c for c in self.__table__.columns]
        for c in _all_bases(type(self)):
            if hasattr(c, "__table__"):
                columns_to_check.extend(c.__table__.columns)

        primary_key_values = {}
        for column in columns_to_check:
            if column.primary_key:
                value = getattr(self, column.name)
                primary_key_values[column] = value

        if all(v is None for v in primary_key_values.values()):
            return None  # special value

        return primary_key_values

    def unique_field_values(self) -> Optional[Tuple[Any]]:
        """
        Returns a tuple of all values that are uniquely constrained (pk or unique).
        Special value None if all of the unique columns are None
        """
        from .utils import _all_bases

        columns_to_check = [c for c in self.__table__.columns]
        for c in _all_bases(type(self)):
            if hasattr(c, "__table__"):
                columns_to_check.extend(c.__table__.columns)

        unique_field_values = []
        for column in columns_to_check:
            # column.primary_key == True for 1+ column guaranteed
            if column.unique or column.primary_key:
                value = getattr(self, column.name)
                unique_field_values.append(value)

        if all(v is None for v in unique_field_values):
            return None  # special value

        return tuple(unique_field_values)

    def merge(self, other):
        """
        Merge column values from other into self, raising an error on conflicts between non-null fields.
        Special handling for JSONB columns where dicts are combined, erroring if overlapping values conflict.
        """
        if self.__class__ != other.__class__:
            raise Exception(
                f"cannot merge {self.__class__} instance with {other.__class__} instance"
            )

        # also need to handle columns for all superclasses
        from .utils import _all_bases

        for column in self.__table__.columns + [
            c
            for b in _all_bases(type(self))
            if hasattr(b, "__table__")
            for c in b.__table__.columns
        ]:
            if hasattr(self, column.name):
                current = getattr(self, column.name)
                incoming = getattr(other, column.name)

                # sqlalchemy unwraps JSONB to dict behind the scenes
                if isinstance(current, dict) and isinstance(incoming, dict):
                    # make a deepcopy to make sure it's a faithful copy
                    from copy import deepcopy

                    old = deepcopy(current)

                    # update and reset the value
                    current.update(incoming)
                    setattr(self, column.name, current)

                    # update incoming with the original, faithful copy from this instance
                    # that way, current and incoming will disagree if any values overlap
                    # do this after we update current to make sure that it's unaffected
                    incoming.update(old)

                if current is None:
                    setattr(self, column.name, incoming)
                elif incoming is not None and current != incoming:
                    pks = {
                        col.name: value for col, value in self.primary_key_map().items()
                    }
                    raise Exception(
                        f"found conflicting values for {self.__class__.__name__} {pks} for {column.name} : {current}!={incoming}"
                    )

    def to_dict(self) -> Dict[str, Any]:
        """Returns a dict of all non-null columns (by name) and their values"""
        # avoid circular imports
        from .utils import _all_bases

        columns_to_check = [c for c in type(self).__table__.columns]
        for b in _all_bases(type(self)):
            if hasattr(b, "__table__"):
                columns_to_check.extend(b.__table__.columns)

        ret = {
            c.name: getattr(self, c.name)
            for c in columns_to_check
            if hasattr(self, c.name)
        }
        ret = {k: v for k, v in ret.items() if v is not None}
        return ret

    @classmethod
    @with_default_session
    def get_by_id(cls, *id, session: Session):
        """
        Returns an instance of this class with given the primary keys if it exists
        Special value None if no instance in the table matches the given values
        """
        with current_app.app_context():
            ret = session.query(cls).get(id)
        return ret



__all__ = [
    "syncall_from_blobs",
    "update_trial_from_metadata_json",
]

from collections import OrderedDict
from copy import deepcopy
from sqlalchemy.orm.session import Session
from typing import Any, Dict, List, Optional, Tuple

from ..models import TrialMetadata, UploadJobs
from .model_core import MetadataModel, with_default_session
from .trial_metadata import (
    ClinicalTrial,
    Cohort,
    CollectionEvent,
    Participant,
    Sample,
    Shipment,
)
from .utils import *
from ...config.logging import get_logger

logger = get_logger(__name__)


@with_default_session
def _make_sample_to_shipment_map(trial_id: str, *, session: Session) -> Dict[str, str]:
    """
    For a given trial, generate a mapping from a cimac_id to the manifest_id it was loaded.
    Uses the UploadJobs where each patch contains only a single shipment and its samples.
    """
    sample_to_shipment_map = {}

    uploads = (
        session.query(UploadJobs)
        .filter(UploadJobs.trial_id == trial_id)
        .order_by(UploadJobs._created)
        .all()
    )
    for upload in uploads:
        shipments = upload.metadata_patch.get("shipments", [])
        if len(shipments) == 0:
            continue
        elif len(shipments) > 1:
            raise Exception(
                f"Multiple shipments in single upload: {upload.id} on trial {upload.trial_id}"
            )
        # else: # len(shipments) == 1

        manifest_id = shipments[0]["manifest_id"]

        for partic in upload.metadata_patch["participants"]:
            for sample in partic["samples"]:
                sample_to_shipment_map[sample["cimac_id"]] = manifest_id

    return sample_to_shipment_map


def update_trial_from_metadata_json(trial_md: dict) -> List[Exception]:
    db_trial = ClinicalTrial.get_by_id(trial_md["protocol_identifier"])
    if db_trial is None:
        raise Exception(
            f"Trial {trial_md['protocol_identifier']} not found in relational tables."
        )

    # relational hook validates by "upsert"
    # also might need to remove cohorts / collection event too
    ordered_records_to_add, records_to_remove = OrderedDict(), []
    ordered_records_to_add[ClinicalTrial] = [
        ClinicalTrial(
            # borrowing from sync_schemas since we're already unmarshalling
            # better handled by a UI change to {key: value}, then old=**request.json
            **_get_all_values(target=ClinicalTrial, old=trial_md)
        )
    ]

    # currently need to compare metadata to look for changes in other tables
    # as UI always returns full list of all values
    old_cohorts = set(db_trial.allowed_cohort_names)
    new_cohorts = set(trial_md.get("allowed_cohort_names", []))
    ordered_records_to_add[Cohort] = [
        Cohort(trial_id=trial_md["protocol_identifier"], cohort_name=value)
        for value in new_cohorts.difference(old_cohorts)
    ]
    records_to_remove.extend(
        Cohort.get_by_id(trial_md["protocol_identifier"], value)
        for value in old_cohorts.difference(new_cohorts)
    )

    old_collections = set(db_trial.allowed_collection_event_names)
    new_collections = set(trial_md.get("allowed_collection_event_names", []))
    ordered_records_to_add[CollectionEvent] = [
        CollectionEvent(trial_id=trial_md["protocol_identifier"], event_name=value)
        for value in new_collections.difference(old_collections)
    ]
    records_to_remove.extend(
        CollectionEvent.get_by_id(trial_md["protocol_identifier"], value)
        for value in old_collections.difference(new_collections)
    )

    return in_single_transaction(
        {
            insert_record_batch: {"ordered_records": ordered_records_to_add},
            remove_record_batch: {"records": records_to_remove},
        }
    )


@with_default_session
def syncall_from_blobs(dry_run: bool = False, *, session: Session) -> List[Exception]:
    """
    Sync all trials from the JSON blobs in TrialMetadata to the new relational tables.
    Handles ClinicalTrial, CollectionEvent, Cohort, Shipment, Participant, and Sample.
    Return a list of all errors encountered.
    """
    errors = []

    trials = session.query(TrialMetadata).all()
    for trial in trials:
        new_errors, new_trial = _generate_new_trial(trial, session=session)
        errors.extend(new_errors)
        if new_trial is None:
            continue

        errors.extend(_sync_collection_events(trial, new_trial))
        errors.extend(_sync_cohorts(trial, new_trial))
        errors.extend(_sync_shipments(trial, new_trial))

        # Since we'll be referencing these for Participant and Sample
        session.flush()

        errors.extend(_sync_participants_and_samples(trial, new_trial))

    # final flush before commit / rollback
    session.flush()

    if dry_run or len(errors):
        session.rollback()
    else:
        session.commit()
    session.close()

    if len(errors):
        logger.error(
            f"Errors in syncall_from_blobs: {len(errors)}\n"
            + "\n".join([str(e) for e in errors])
        )
    else:
        logger.info("No errors in syncall_from_blobs")

    return errors


def _get_all_values(
    target: MetadataModel, old: dict, drop: List[str] = []
) -> Dict[str, Any]:
    """Returns all of the values from `old` that are columns of `target` excepting anything keys in `drop`"""
    columns_to_check = [c for c in target.__table__.columns]
    for b in _all_bases(type(target)):
        if hasattr(b, "__table__"):
            columns_to_check.extend(b.__table__.columns)

    ret = {
        c.name: old[c.name]
        for c in columns_to_check
        if c.name not in drop and c.name in old
    }

    # put all original data into the json_data column
    if "json_data" not in drop and any(c.name == "json_data" for c in columns_to_check):
        ret["json_data"] = deepcopy(old)

    return ret


def _generate_new_trial(
    trial: TrialMetadata, *, session: Session
) -> Tuple[List[Exception], Optional[ClinicalTrial]]:
    """
    Given a TrialMetadata, set up a new ClinicalTrial.
    Returns one of:
        [Exception()], None
        [], ClinicalTrial()
    """
    try:
        new_trial = session.merge(
            ClinicalTrial(
                protocol_identifier=trial.trial_id,
                **_get_all_values(
                    target=ClinicalTrial,
                    old=trial.metadata_json,
                    drop=["protocol_identifier"],
                ),
            )
        )
        session.flush()
        return [], new_trial

    except Exception as e:
        return [e], None


def _sync_collection_events(
    trial: TrialMetadata, new_trial: ClinicalTrial
) -> List[Exception]:
    """
    Given corresponding TrialMetadata and ClinicalTrial instances, copy over collection events.
    In JSON, stored as list[string] at /allowed_collection_event_names.
    """
    records = {
        CollectionEvent: [
            CollectionEvent(
                event_name=event_name, trial_id=new_trial.protocol_identifier
            )
            for event_name in trial.metadata_json["allowed_collection_event_names"]
        ]
    }

    return insert_record_batch(records, hold_commit=True)


def _sync_cohorts(trial: TrialMetadata, new_trial: ClinicalTrial) -> List[Exception]:
    """
    Given corresponding TrialMetadata and ClinicalTrial instances, copy over cohorts.
    In JSON, stored as list[string] at /allowed_cohort_names.
    """
    records = {
        Cohort: [
            Cohort(cohort_name=cohort_name, trial_id=new_trial.protocol_identifier)
            for cohort_name in trial.metadata_json["allowed_cohort_names"]
        ]
    }

    return insert_record_batch(records, hold_commit=True)


def _sync_shipments(trial: TrialMetadata, new_trial: ClinicalTrial) -> List[Exception]:
    """
    Given corresponding TrialMetadata and ClinicalTrial instances, copy over shipments.
    In JSON, stored as list[shipment] at /shipments, where each shipment is dict[<column name> : <value>]
    """
    records = {
        Shipment: [
            Shipment(
                trial_id=new_trial.protocol_identifier,
                **_get_all_values(target=Shipment, old=shipment),
            )
            for shipment in trial.metadata_json.get("shipments", [])
        ]
    }

    return insert_record_batch(records, hold_commit=True)


def _sync_participants_and_samples(
    trial: TrialMetadata,
    new_trial: ClinicalTrial,
) -> List[Exception]:
    """
    Given corresponding TrialMetadata and ClinicalTrial instances, copy over participants and samples.
    In JSON, stored as list[participant] at /participants, where each participant is dict[<column name> : <value>]
    with /participants/*/samples : list[sample] where each sample is dict[<column name> : <value>].

    For Participant, JSON participant_id becomes relational trial_participant_id for clarity.
    """
    errors = []
    shipment_map = _make_sample_to_shipment_map(trial.trial_id)

    participants = trial.metadata_json["participants"]
    for partic in participants:
        # special handling for name change
        partic["trial_participant_id"] = partic.pop("participant_id")

        records = {
            Participant: [
                Participant(
                    trial_id=new_trial.protocol_identifier,
                    **_get_all_values(target=Participant, old=partic),
                )
            ]
        }

        records[Sample] = []
        samples = partic.get("samples", [])
        for sample in samples:
            if sample["cimac_id"] not in shipment_map:
                errors.append(
                    Exception(f"No manifest_id found for sample {sample['cimac_id']}")
                )
                continue

            records[Sample].append(
                Sample(
                    trial_id=new_trial.protocol_identifier,
                    manifest_id=shipment_map[sample["cimac_id"]],
                    # neither of the above are in sample, so don't need a `drop`
                    **_get_all_values(target=Sample, old=sample),
                )
            )

        errors.extend(insert_record_batch(records, hold_commit=True))

    return errors



__all__ = [
    "HandeAssay",
    "WesBamAssay",
    "WesFastqAssay",
]

from .core import Entry, MetadataTemplate, WorksheetConfig
from .model_core import get_property

from .assay_metadata import (
    HandeImage,
    HandeRecord,
    HandeUpload,
    WESRecord,
    WESUpload,
)
from .file_metadata import (
    BamFile,
    Fastq_gzFile,
    NGSAssayFiles,
)


HandeAssay = MetadataTemplate(
    upload_type="hande",
    purpose="assay",
    worksheet_configs=[
        WorksheetConfig(
            "H&E",
            [
                Entry(HandeUpload.trial_id, name="Protocol identifier"),
                Entry(HandeUpload.assay_creator),
            ],
            {
                "Samples": [
                    Entry(HandeRecord.cimac_id),
                    Entry(
                        HandeImage.local_path,
                        name="Image file",
                        gcs_uri_format="{trial_id}/hande/{cimac_id}/image_file.svs",
                        process_as={HandeRecord.image_url: get_property("object_url")},
                    ),
                    Entry(
                        HandeRecord.tumor_tissue_percentage,
                        name="Tumor tissue (% total area)",
                    ),
                    Entry(
                        HandeRecord.viable_tumor_percentage,
                        name="Viable tumor (% area)",
                    ),
                    Entry(
                        HandeRecord.viable_stroma_percentage,
                        name="Viable stroma (% area)",
                    ),
                    Entry(
                        HandeRecord.necrosis_percentage,
                        name="Necrosis (% area)",
                    ),
                    Entry(
                        HandeRecord.fibrosis_percentage,
                        name="Fibrosis (% area)",
                    ),
                    Entry(HandeRecord.comment),
                ]
            },
        )
    ],
    constants={
        HandeUpload.upload_type: "hande",
        HandeUpload.multifile: True,
    },
)

WesFastqAssay = MetadataTemplate(
    upload_type="wes",
    purpose="assay",
    worksheet_configs=[
        WorksheetConfig(
            "WES",
            [
                Entry(WESUpload.trial_id, name="Protocol identifier"),
                Entry(WESUpload.assay_creator),
                Entry(WESUpload.sequencing_protocol),
                Entry(WESUpload.library_kit),
                Entry(WESUpload.sequencer_platform),
                Entry(WESUpload.paired_end_reads),
                Entry(WESUpload.read_length),
                Entry(WESUpload.bait_set),
            ],
            {
                "Samples": [
                    Entry(WESRecord.cimac_id),
                    Entry(NGSAssayFiles.lane),
                    Entry(
                        Fastq_gzFile.local_path,
                        name="Forward fastq",
                        gcs_uri_format="{trial_id}/wes/{cimac_id}/r1_L{lane}.fastq.gz",
                        process_as={
                            NGSAssayFiles.r1_object_url: get_property("object_url"),
                        },
                    ),
                    Entry(
                        Fastq_gzFile.local_path,
                        name="Reverse fastq",
                        gcs_uri_format="{trial_id}/wes/{cimac_id}/r2_L{lane}.fastq.gz",
                        process_as={
                            NGSAssayFiles.r2_object_url: get_property("object_url"),
                        },
                    ),
                    Entry(WESRecord.sequencing_date),
                    Entry(WESRecord.quality_flag),
                ]
            },
        )
    ],
    constants={
        WESUpload.upload_type: "wes",
        WESUpload.multifile: True,
    },
)

WesBamAssay = MetadataTemplate(
    upload_type="wes",
    purpose="assay",
    worksheet_configs=[
        WorksheetConfig(
            "WES",
            [
                Entry(WESUpload.trial_id, name="Protocol identifier"),
                Entry(WESUpload.assay_creator),
                Entry(WESUpload.sequencing_protocol),
                Entry(WESUpload.library_kit),
                Entry(WESUpload.sequencer_platform),
                Entry(WESUpload.paired_end_reads),
                Entry(WESUpload.read_length),
                Entry(WESUpload.bait_set),
            ],
            {
                "Samples": [
                    Entry(WESRecord.cimac_id),
                    Entry(NGSAssayFiles.number),
                    Entry(
                        BamFile.local_path,
                        name="Bam file",
                        gcs_uri_format="{trial_id}/wes/{cimac_id}/reads_{number}.bam",
                        process_as={
                            NGSAssayFiles.bam_object_url: get_property("object_url"),
                        },
                    ),
                    Entry(WESRecord.sequencing_date),
                    Entry(WESRecord.quality_flag),
                ]
            },
        )
    ],
    constants={
        WESUpload.upload_type: "wes",
        WESUpload.multifile: True,
    },
)


if __name__ == "__main__":
    WesFastqAssay.write("tests/models/templates/examples/wes_fastq_assay.xlsx")
    WesBamAssay.write("tests/models/templates/examples/wes_bam_assay.xlsx")



# __all__ = [
#     "ClinicalData",
#     "ClinicalYesNoEnum",
#     "DiseaseStageEnum",
#     "DiseaseGradeEnum",
#     "EcogPsEnum",
#     "EthnicityEnum",
#     "GenderEnum",
#     "ModAnnArborEnum",
#     "MostRecentTreatmentEnum",
#     "PriorTherapyTypeEnum",
#     "RaceEnum",
#     "RecentTreatmentResponseEnum",
# ]

# from sqlalchemy import (
#     CheckConstraint,
#     Column,
#     Enum,
#     ForeignKeyConstraint,
#     Integer,
#     Numeric,
#     String,
# )

# from .model_core import MetadataModel
# from .trial_metadata import Participant

# RaceEnum = Enum(
#     "White",
#     "American Indian or Alaska Native",
#     "Black or African American",
#     "Asian",
#     "Native Hawaiian or Other Pacific Islander",
#     "Other",
#     "Unknown",
#     "Not reported",
#     "Not allowed to collect",
#     name="race_enum",
# )
# GenderEnum = Enum(
#     "Female", "Male", "Unknown", "Unspecified", "Not reported", name="gender_enum",
# )
# EthnicityEnum = Enum(
#     "Hispanic or Latino",
#     "Not Hispanic or Latino",
#     "Unknown",
#     "Not reported",
#     "Not allowed to collect",
#     name="ethnicity_enum",
# )
# ClinicalYesNoEnum = Enum(
#     "Yes", "No", "Unknown", "Not applicable", name="clinical_yes_no_enum",
# )
# PriorTherapyTypeEnum = Enum(
#     "Anti-Retroviral Therapy",
#     "Antisense",
#     "Bone Marrow Transplant",
#     "Chemotherapy not otherwise specified (NOS)",
#     "Chemotherapy multiple agents systemic",
#     "Chemotherapy non-cytotoxic",
#     "Chemotherapy single agent systemic",
#     "Drug and/or immunotherapy",
#     "Gene Transfer",
#     "Hematopoietic stem cell",
#     "Hormonal Therapy",
#     "Image Directed local therapy",
#     "No prior therapy",
#     "Oncolytic virotherapy",
#     "Prior therapy not otherwise specified (NOS)",
#     "Radiation Therapy",
#     "Surgery",
#     "Therapy (NOS)",
#     "Vaccine",
#     name="prior_therapy_type_enum",
# )

# ModAnnArborEnum = Enum(
#     "I", "II", "III", "IV", "Unknown", "Not Applicable", name="mod_ann_arbor_enum",
# )
# EcogPsEnum = Enum(
#     "0",
#     "1",
#     "2",
#     "3",
#     "4",
#     "5",
#     "Data Missing",
#     "Unknown",
#     "Not Applicable",
#     "Not Reported",
#     name="ecog_ps_enum",
# )
# MostRecentTreatmentEnum = Enum(
#     "ABVD",
#     "BEACOPP",
#     "MOPP",
#     "ICE",
#     "Gemcitabine",
#     "Vinblastine",
#     "Bendamustine",
#     "Brentuximab Vedotin",
#     "Lenalidomide (Revlimid)",
#     "RAD001 (Everolimus)",
#     "Rituximab",
#     "Autologous Stem Cell Transplant",
#     "Allogeneic Stem Cell Transplant",
#     "Other - Experimental Targeted",
#     "Therapy",
#     "Other - Immunotherapy",
#     "Other - not specified",
#     name="most_recent_treatment_enum",
# )
# RecentTreatmentResponseEnum = Enum(
#     "Complete Remission",
#     "Partial Remission",
#     "Stable Disease",
#     "Progressive Disease",
#     "Unevaluable",
#     "Not Reported",
#     "Unknown",
#     name="recent_treatment_response_enum",
# )
# DiseaseStageEnum = Enum(
#     "0",
#     "0a",
#     "0is",
#     "I",
#     "IA",
#     "IA1",
#     "IA2",
#     "IB",
#     "IB1",
#     "IB2",
#     "IC",
#     "II",
#     "IIA",
#     "IIA1",
#     "IIA2",
#     "IIB",
#     "IIC",
#     "IIC1",
#     "III",
#     "IIIA",
#     "IIIB",
#     "IIIC",
#     "IIIC1",
#     "IIIC2",
#     "IS",
#     "IV",
#     "IVA",
#     "IVB",
#     "IVC",
#     "Tis",
#     "X",
#     "Unknown",
#     "Not Reported",
#     name="disease_stage_enum",
# )
# DiseaseGradeEnum = Enum(
#     "G1",
#     "G2",
#     "G3",
#     "G4",
#     "GX",
#     "GB",
#     "High Grade",
#     "Low Grade",
#     "Unknown",
#     "Not Reported",
#     "Not Applicable",
#     "Data Missing",
#     name="disease_grade_enum",
# )


# class ClinicalData(MetadataModel):
#     __tablename__ = "clinical_data"

#     trial_id = Column(String, primary_key=True)
#     cimac_participant_id = Column(String, primary_key=True)

#     __table_args__ = (
#         ForeignKeyConstraint(
#             [trial_id, cimac_participant_id],
#             [Participant.trial_id, Participant.cimac_participant_id],
#         ),
#     )

#     # demographics
#     race = Column(
#         RaceEnum,
#         doc="An arbitrary classification based on physical characteristics; a group of persons related by common descent or heredity (U.S. Center for Disease Control).",
#     )
#     gender = Column(GenderEnum, doc="Sex of the participant.",)
#     ethnicity = Column(EthnicityEnum, doc="Ethnicity of the participant.",)
#     age = Column(
#         Numeric,
#         CheckConstraint("age >= 0"),
#         doc="The age of the subject expressed in years",
#     )

#     # history
#     prior_surgery = Column(
#         ClinicalYesNoEnum,
#         doc="An indication whether or not there was any surgery history to report.",
#     )
#     prior_radiation_therapy = Column(
#         ClinicalYesNoEnum,
#         doc="An indication whether or not there was any radiation therapy history to report.",
#     )
#     prior_immunotherapy = Column(
#         ClinicalYesNoEnum,
#         doc="An indication whether or not there was any immunotherapy history to report.",
#     )
#     number_prior_systemic_treatments = Column(
#         Integer,
#         CheckConstraint("number_prior_systemic_treatments >= 0"),
#         doc="Total number of patient's prior systemic treatments, if any.",
#     )
#     prior_therapy_type = Column(
#         PriorTherapyTypeEnum,
#         doc="Text term that describes the kind of treatment administered.",
#     )

#     # disease & baseline
#     mod_ann_arbor_stage = Column(
#         ModAnnArborEnum,
#         doc="Text term that represents the clinical stage for lymphoma using the Ann Arbor Lymphoma Staging System.",
#     )
#     ecog_ps = Column(
#         EcogPsEnum,
#         doc="The ECOG functional performance status of the patient/participant.",
#     )
#     years_from_initial_diagnosis = Column(
#         Numeric,
#         CheckConstraint("years_from_initial_diagnosis >= 0"),
#         doc="Time at enrollment since initial diagnosis, in years.",
#     )
#     type_of_most_recent_treatment = Column(
#         MostRecentTreatmentEnum,
#         doc="Type of most recent treatment administered to the patient.",
#     )
#     response_to_most_recent_treatment = Column(
#         RecentTreatmentResponseEnum,
#         doc="Type of most recent treatment administered to the patient.",
#     )
#     duration_of_remission = Column(
#         Numeric,
#         CheckConstraint("duration_of_remission >= 0"),
#         doc="Duration of remission, in months.",
#     )
#     years_from_recent_treatment = Column(
#         Numeric,
#         CheckConstraint("years_from_recent_treatment >= 0"),
#         doc="Time at enrollment since most recent treatment, in years.",
#     )
#     disease_stage = Column(
#         DiseaseStageEnum,
#         doc="Neoplasm American Joint Committee on Cancer Clinical Group Stage. Stage group determined from clinical information on the tumor (T), regional node (N) and metastases (M) and by grouping cases with similar prognosis for cancer.",
#     )
#     disease_grade = Column(
#         DiseaseGradeEnum,
#         doc="Numeric value to express the degree of abnormality of cancer cells, a measure of differentiation and aggressiveness.",
#     )



__all__ = [
    # "ClinicalDataTemplate",
    "PbmcManifest",
    "TissueSlideManifest",
]

from sqlalchemy import Column
from typing import Any, Dict, List

from .core import Entry, MetadataTemplate, WorksheetConfig
from .model_core import cimac_id_to_cimac_participant_id, identity

from .trial_metadata import Participant, Sample, Shipment

# from .extra_metadata import ClinicalData


# ClinicalDataTemplate = MetadataTemplate(
#     upload_type="clinical_data",
#     purpose="manifest",
#     worksheet_configs=[
#         WorksheetConfig(
#             "Tier 1",
#             [Entry(ClinicalData.trial_id, name="Protocol identifier")],
#             {
#                 "Demographics": [
#                     Entry(
#                         Participant.cimac_participant_id,
#                         process_as={ClinicalData.cimac_participant_id: identity},
#                     ),
#                     Entry(ClinicalData.race),
#                     Entry(ClinicalData.gender, name="Sex",),
#                     Entry(ClinicalData.ethnicity, name="Ethnic group"),
#                     Entry(ClinicalData.age),
#                 ],
#                 "History": [
#                     Entry(ClinicalData.prior_surgery),
#                     Entry(ClinicalData.prior_radiation_therapy),
#                     Entry(ClinicalData.prior_immunotherapy),
#                     Entry(ClinicalData.number_prior_systemic_treatments),
#                     Entry(ClinicalData.prior_therapy_type),
#                 ],
#                 "Disease & Baseline": [
#                     Entry(
#                         ClinicalData.mod_ann_arbor_stage,
#                         name="Modified Ann Arbor Stage",
#                     ),
#                     Entry(ClinicalData.ecog_ps),
#                     Entry(ClinicalData.years_from_initial_diagnosis),
#                     Entry(ClinicalData.type_of_most_recent_treatment),
#                     Entry(ClinicalData.response_to_most_recent_treatment),
#                     Entry(ClinicalData.duration_of_remission),
#                     Entry(ClinicalData.years_from_recent_treatment),
#                     Entry(ClinicalData.disease_stage),
#                     Entry(ClinicalData.disease_grade),
#                 ],
#             },
#         ),
#     ],
# )

# To share the base WorksheetConfig's between all of the templates
# we'll use variables to represent where we can add custom elements
# into the otherwise static configs and then return the object
# # not using a subclass so that they are still MetadataTemplate's
# PREFERED: changes may need to be made here to allow for new custom elements
# ALERNATE: can add as post-hoc modifications to individual manifest templates
def _BaseManifestTemplate(
    upload_type: str,
    filled_by_biorepository: List[Entry],
    filled_by_cimac_lab: List[Entry],
    essential_patient_data: bool = False,
    constants: Dict[Column, Any] = {},
):
    worksheets = [
        WorksheetConfig(
            "Shipment",
            [
                Entry(
                    Shipment.trial_id,
                    name="Protocol identifier",
                    process_as={
                        Participant.trial_id: identity,
                        Sample.trial_id: identity,
                    },
                ),
                Entry(
                    Shipment.manifest_id,
                    process_as={Sample.manifest_id: identity},
                ),
                Entry(Shipment.assay_priority, deprecated=True),
                Entry(
                    Shipment.assay_type,
                    process_as={Sample.intended_assay: identity},
                    deprecated=True,
                ),
                Entry(Shipment.receiving_party, deprecated=True),
                Entry(Shipment.courier, deprecated=True),
                Entry(Shipment.tracking_number, deprecated=True),
                Entry(Shipment.account_number, deprecated=True),
                Entry(Shipment.shipping_condition, deprecated=True),
                Entry(Shipment.date_shipped, deprecated=True),
                Entry(Shipment.date_received, deprecated=True),
                Entry(Shipment.quality_of_shipment, deprecated=True),
                Entry(Shipment.ship_from, deprecated=True),
                Entry(Shipment.ship_to, deprecated=True),
            ],
            {},
        ),
        WorksheetConfig(
            "Samples",
            [],
            {
                "IDs": [
                    Entry(
                        Sample.shipping_entry_number, name="Entry (#)", deprecated=True
                    ),
                    Entry(Sample.collection_event_name),
                    Entry(Participant.cohort_name),
                    Entry(
                        Participant.trial_participant_id,
                        name="Participant id",
                        deprecated=True,
                    ),
                    Entry(Sample.parent_sample_id, deprecated=True),
                    Entry(Sample.processed_sample_id, deprecated=True),
                    Entry(
                        Sample.cimac_id,
                        process_as={
                            Sample.cimac_participant_id: cimac_id_to_cimac_participant_id,
                            Participant.cimac_participant_id: cimac_id_to_cimac_participant_id,
                        },
                    ),
                ],
                "Filled by Biorepository": filled_by_biorepository,
                "Filled by CIMAC Lab": filled_by_cimac_lab,
            },
        ),
    ]
    if essential_patient_data:
        worksheets.append(
            WorksheetConfig(
                "Essential Patient Data",
                [],
                {
                    "Path Concordance Verification": [
                        Entry(
                            Sample.shipping_entry_number,
                            name="Entry (#)",
                            deprecated=True,
                        ),
                        Entry(
                            Participant.trial_participant_id,
                            name="Participant id",
                            deprecated=True,
                        ),
                        Entry(
                            Sample.cimac_id,
                            process_as={
                                Participant.cimac_participant_id: cimac_id_to_cimac_participant_id,
                                Sample.cimac_participant_id: cimac_id_to_cimac_participant_id,
                            },
                        ),
                        Entry(Sample.surgical_pathology_report_id, deprecated=True),
                        Entry(Sample.clinical_report_id, deprecated=True),
                        Entry(Sample.collection_event_name),
                        Entry(Sample.diagnosis_verification, deprecated=True),
                    ],
                    "ICD-0-3 Code/Description": [
                        Entry(Sample.site_description, deprecated=True),
                        Entry(Sample.topography_code, deprecated=True),
                        Entry(Sample.topography_description, deprecated=True),
                        Entry(
                            Sample.histology_behavior,
                            name="Histology/behavior",
                            deprecated=True,
                        ),
                        Entry(
                            Sample.histology_behavior_description,
                            name="Histology/behavior description",
                            deprecated=True,
                        ),
                    ],
                    "Demographics": [
                        Entry(Participant.gender, name="Sex", deprecated=True),
                        Entry(Participant.race, deprecated=True),
                        Entry(
                            Participant.ethnicity, name="Ethnic group", deprecated=True
                        ),
                    ],
                },
            )
        )
    return MetadataTemplate(
        upload_type=upload_type,
        purpose="manifest",
        worksheet_configs=worksheets,
        constants=constants,
    )


PbmcManifest = _BaseManifestTemplate(
    upload_type="pbmc",
    filled_by_biorepository=[
        Entry(Sample.box_number, deprecated=True),
        Entry(Sample.sample_location, deprecated=True),
        Entry(Sample.type_of_sample, deprecated=True),
        Entry(Sample.sample_collection_procedure, deprecated=True),
        Entry(Sample.type_of_primary_container, deprecated=True),
        Entry(Sample.processed_sample_type, deprecated=True),
        Entry(Sample.processed_sample_volume, deprecated=True),
        Entry(Sample.processed_sample_volume_units, deprecated=True),
        Entry(Sample.processed_sample_concentration, deprecated=True),
        Entry(Sample.processed_sample_concentration_units, deprecated=True),
        Entry(Sample.processed_sample_derivative, deprecated=True),
        Entry(Sample.sample_derivative_volume, deprecated=True),
        Entry(Sample.sample_derivative_volume_units, deprecated=True),
        Entry(Sample.sample_derivative_concentration, deprecated=True),
        Entry(Sample.sample_derivative_concentration_units, deprecated=True),
    ],
    filled_by_cimac_lab=[
        Entry(Sample.pbmc_viability, deprecated=True),
        Entry(Sample.pbmc_recovery, deprecated=True),
        Entry(Sample.pbmc_resting_period_used, deprecated=True),
        Entry(Sample.material_used, deprecated=True),
        Entry(Sample.material_remaining, deprecated=True),
        Entry(Sample.material_storage_condition, deprecated=True),
        Entry(Sample.quality_of_sample, deprecated=True),
        Entry(Sample.sample_replacement, deprecated=True),
        Entry(Sample.residual_sample_use, deprecated=True),
        Entry(Sample.comments, deprecated=True),
    ],
    essential_patient_data=True,
)

TissueSlideManifest = _BaseManifestTemplate(
    upload_type="tissue_slide",
    filled_by_biorepository=[
        Entry(Sample.box_number, deprecated=True),
        Entry(Sample.sample_location, deprecated=True),
        Entry(Sample.type_of_sample, deprecated=True),
        Entry(Sample.type_of_tumor_sample, deprecated=True),
        Entry(Sample.sample_collection_procedure, deprecated=True),
        Entry(Sample.core_number, deprecated=True),
        Entry(
            Sample.fixation_stabilization_type,
            name="Fixation/stabilization type",
            deprecated=True,
        ),
        Entry(Sample.processed_sample_type, deprecated=True),
        Entry(Sample.processed_sample_quantity, deprecated=True),
    ],
    filled_by_cimac_lab=[
        Entry(Sample.material_used, deprecated=True),
        Entry(Sample.material_used_units, deprecated=True),
        Entry(Sample.material_remaining, deprecated=True),
        Entry(Sample.material_remaining_units, deprecated=True),
        Entry(Sample.material_storage_condition, deprecated=True),
        Entry(Sample.quality_of_sample, deprecated=True),
        Entry(Sample.sample_replacement, deprecated=True),
        Entry(Sample.residual_sample_use, deprecated=True),
        Entry(Sample.comments, deprecated=True),
    ],
    essential_patient_data=False,
)



from .trial_metadata import *
from .file_metadata import *
from .assay_metadata import *

# Everything has to be imported after the tables they reference
from .assay_templates import *
from .manifest_templates import *
from .sync_schemas import *

from .model_core import *
from .core import *
from .utils import *

# This maps from full_template_name prism/CLI-compatible
# assay names to the MetadataTemplate instances that
# handles each type of upload.
TEMPLATE_MAP = {
    get_full_template_name(name): value
    for name, value in dict(globals(), **locals()).items()
    if isinstance(value, MetadataTemplate)
}



__all__ = [
    "Entry",
    "ExcelStyles",
    "MetadataTemplate",
    "MODEL_INSERTION_ORDER",
    "row_type_from_string",
    "RowType",
    "WorksheetConfig",
]

import os

os.environ["TZ"] = "UTC"
import datetime
from collections import defaultdict, OrderedDict
from enum import Enum
from warnings import filterwarnings
from typing import (
    Any,
    BinaryIO,
    Dict,
    Callable,
    List,
    Optional,
    OrderedDict as OrderedDict_Type,
    Tuple,
    Type,
    Union,
)

import openpyxl
from sqlalchemy import Column, Enum as SqlEnum
import xlsxwriter
from xlsxwriter.utility import xl_rowcol_to_cell, xl_range

from .model_core import MetadataModel
from .utils import _all_bases, _get_global_insertion_order, insert_record_batch

MODEL_INSERTION_ORDER = _get_global_insertion_order()

filterwarnings(
    action="ignore",
    category=UserWarning,
    message="Data Validation extension is not supported",
    module="openpyxl",
)


def _format_for_json_serialization(value):
    if isinstance(value, datetime.date):
        return value.strftime("%Y%m%d")
    elif isinstance(value, datetime.time):
        return value.strftime("T%H%M%S.%f")
    elif isinstance(value, datetime.datetime):
        return value.strftime("%Y%m%dT%H%M%S.%f")
    else:
        return str(value)


class Entry:
    """
    One field in a metadata worksheet. Provides configuration for reading a
    value from a metadata spreadsheet cell into a sqlalchemy model attribute
    (or set of related attributes).

    Args:
      column: column attribute on a SQLAlchemy model class.
      name: optional human-readable label for this field, if not inferrable from `column`.
      gcs_uri_format: allowed iff hasattr(column.class_, 'object_url'), fstring .
      process_as: optional dictionary mapping column attributes to data processing function.
      deprecated: bool (default False) of whether to add this name/value to the neighboring json_data.

    TO IMPLEMENT:
      encrypt: whether the spreadsheet value should be encrypted before storage.
    """

    def __init__(
        self,
        column: Column,
        name: Optional[str] = None,
        *,
        gcs_uri_format: Optional[str] = None,
        process_as: Dict[Column, Callable[[str], Any]] = {},
        deprecated: bool = False,
    ):
        self.column = column
        self.name = name if name else column.name.replace("_", " ").capitalize()
        self.gcs_uri_format = gcs_uri_format
        self.process_as = process_as
        self.deprecated = deprecated

        if gcs_uri_format and not hasattr(column.class_, "object_url"):
            raise Exception(
                f"gcs_uri_format should only be defined for columns on a File, not {column.class_.__name__}"
            )
        if deprecated:
            if not hasattr(column.class_, "json_data"):
                raise Exception(
                    f"deprecated should only be True for columns with a neighboring json_data, not {column.class_.__name__}"
                )
            elif process_as:
                for col in process_as.keys():
                    if not hasattr(col.class_, "json_data"):
                        raise Exception(
                            f"deprecated should only be True for columns with a neighboring json_data, not {col.class_.__name__} (in process_as)"
                        )

        self.doc = column.doc
        self.sqltype = column.type
        self.enums = self.sqltype.enums if isinstance(self.sqltype, SqlEnum) else None
        self.pytype = self.sqltype.python_type

    def get_column_mapping(self, value, context: dict = {}) -> Dict[Column, Any]:
        """
        Given a value and a surround context to evalutate it in, returns a dict
        pointing table columns to their correctly-typed values. A single item
        if there's no process_as defined for this entry.
        Special handling for missing required values, date/time types, GCS URI
        formatting. Handling process_as is handled last.
        """
        if value is None:
            if self.column.nullable:
                return {self.column: None}
            else:
                raise Exception(f"Missing required value {self.name}")

        try:
            # Handle date/time parsing funkiness
            if self.pytype == datetime.time:
                try:
                    if not isinstance(value, datetime.datetime):
                        value = openpyxl.utils.datetime.from_excel(value)
                    processed_value = value.time()
                except:
                    raise TypeError(f"{value} is not a valid time")
            elif self.pytype == datetime.date:
                try:
                    if not isinstance(value, datetime.datetime):
                        value = openpyxl.utils.datetime.from_excel(value)
                    processed_value = value.date()
                except:
                    raise TypeError(f"{value} is not a valid date")

            else:
                processed_value = self.pytype(value)

            column_mapping = {self.column: processed_value}

            # Handle GCS URI formatting
            if self.gcs_uri_format:
                format_dict = {k.name: v for k, v in column_mapping.items()}
                format_dict.update({k.name: v for k, v in context.items()})
                column_mapping[
                    getattr(self.column.class_, "object_url")
                ] = self.gcs_uri_format.format(**format_dict)

            # Handle deprecation to add to json_data column
            # class_."json_data" is JSONB -> dict
            # specific handling for dict in MetadataModel.merge and MetadataTemplate.read
            if self.deprecated:
                column_mapping[getattr(self.column.class_, "json_data")] = {
                    self.column.name: _format_for_json_serialization(value)
                }

            # Finally, handle process_as
            for column, process in self.process_as.items():
                context.update(column_mapping)
                column_mapping[column] = process(value, context)

                if self.deprecated:
                    column_mapping[getattr(column.class_, "json_data")] = {
                        column.name: _format_for_json_serialization(value)
                    }

        except Exception as e:
            raise Exception(
                f"Error in processing {self.name}={value}({type(value)}) as {self.pytype}: {e}"
            ) from e
        return column_mapping


class WorksheetConfig:
    """
    A worksheet within a metadata spreadsheet. A worksheet has a name, a list
    of `preamble_rows` and a dictionary mapping data column group headers to lists
    of `data_sections`.
    """

    name: str
    preamble: List[Entry]
    data_sections: Dict[str, List[Entry]]

    def __init__(
        self, name: str, preamble: List[Entry], data_sections: Dict[str, List[Entry]]
    ):
        self.name = name
        self.preamble = preamble
        self.data_sections = data_sections


class RowType(Enum):
    """Annotations denoting what type of data a template row contains."""

    TITLE = "#title"
    SKIP = "#skip"
    HEADER = "#header"
    PREAMBLE = "#preamble"
    DATA = "#data"


def row_type_from_string(maybe_type: str) -> Optional[RowType]:
    """Returns None if the input is not a valid RowType."""
    try:
        return RowType(maybe_type)
    except ValueError:
        return None


def _format_validation_range(
    validation_rows, validation_column, data_dict_worksheet_name
):
    start = xl_rowcol_to_cell(
        1,  # 1 is to skip first row in DD sheet that is for header
        validation_column,
        row_abs=True,
        col_abs=True,
    )
    stop = xl_rowcol_to_cell(
        validation_rows, validation_column, row_abs=True, col_abs=True
    )

    return f"'{data_dict_worksheet_name}'!{start}:{stop}"


class ExcelStyles:
    COLUMN_WIDTH = 30

    TITLE_STYLE_PROPS = {
        "border": 1,
        "bg_color": "#ffffb3",
        "bold": True,
        "align": "center",
        "text_wrap": True,
        "valign": "vcenter",
        "indent": 1,
    }
    PREAMBLE_STYLE_PROPS = {
        "border": 0,
        "top": 2,
        "top_color": "white",
        "bottom": 2,
        "bottom_color": "white",
        "bg_color": "#b2d2f6",
        "bold": True,
        "align": "right",
        "text_wrap": True,
        "valign": "vcenter",
        "indent": 1,
    }
    HEADER_STYLE_PROPS = {
        "border": 1,
        "bg_color": "#C6EFCE",
        "bold": True,
        "align": "center",
        "valign": "vcenter",
        "indent": 1,
    }
    DATA_STYLE_PROPS = {
        "border": 1,
        "bg_color": "#5fa3f0",
        "bold": True,
        "align": "center",
        "text_wrap": True,
        "valign": "vcenter",
        "indent": 1,
    }
    DIRECTIVE_STYLE_PROPS = {
        "border": 1,
        "bg_color": "#ffffb3",
        "bold": True,
        "align": "center",
        "text_wrap": True,
        "valign": "vcenter",
        "indent": 1,
    }
    COMMENT_STYLE_PROPS = {
        "color": "white",
        "font_size": 10,
        "x_scale": 2,
        "author": "CIDC",
    }

    def __init__(self, workbook: xlsxwriter.Workbook):
        self.TITLE_STYLE = workbook.add_format(self.TITLE_STYLE_PROPS)
        self.PREAMBLE_STYLE = workbook.add_format(self.PREAMBLE_STYLE_PROPS)
        self.HEADER_STYLE = workbook.add_format(self.HEADER_STYLE_PROPS)
        self.DATA_STYLE = workbook.add_format(self.DATA_STYLE_PROPS)
        self.DIRECTIVE_STYLE = workbook.add_format(self.DIRECTIVE_STYLE_PROPS)


class MetadataTemplate:
    """
    A metadata template. Must have attributes `upload_type`, `purpose`, and `worksheet_configs` defined.
    An optional `constants` dict allows for hidden values to be used later in reading the data, but does not
    affect the output template XLSX from `.write()`
    """

    upload_type: str
    purpose: str
    worksheet_configs: List[WorksheetConfig]
    constants: Dict[Column, Any]

    DATA_ROWS = 2000
    DATA_DICT_SHEETNAME = "Data Dictionary"

    def __init__(
        self,
        upload_type: str,
        purpose: str,
        worksheet_configs: List[WorksheetConfig],
        constants: Dict[Column, Any] = {},
    ):
        self.upload_type = upload_type
        self.purpose = purpose
        self.worksheet_configs = worksheet_configs
        self.constants = constants

    def read_and_insert(self, filename: Union[str, BinaryIO]) -> List[Exception]:
        """
        Extract the models from a populated instance of this template and try to
        insert them, rolling back and returning a list of errors if any are encountered.
        """
        try:
            records = self.read(filename)
        except Exception as e:
            return [e]
        else:
            return insert_record_batch(records)

    def read(
        self, filename: Union[str, BinaryIO]
    ) -> OrderedDict_Type[Type, List[MetadataModel]]:
        """
        Extract a list of SQLAlchemy models in insertion order from a populated
        instance of this template.
        """
        workbook = openpyxl.load_workbook(filename)

        # The preamble values flow across all sheets for context
        # (such as trial_id) to only need to collect each value once.
        # Constants also are used for all sheets.
        preamble_dict = self.constants.copy()

        # Extract partial model instances from the template
        model_instances: List[MetadataModel] = []
        model_dicts: List[Dict[Column, Any]] = []
        for config in self.worksheet_configs:
            if config.name not in workbook:
                raise Exception(
                    f"Missing expected worksheet {config.name}, please include even if all fields are optional"
                )

            # split the preamble and data rows
            preamble_rows = []
            data_rows = []
            data_header = None
            for n, row in enumerate(workbook[config.name].iter_rows()):
                # if no entries, skip it
                if all(cell.value is None for cell in row[1:]):
                    continue
                # row[0] is the type
                row_type: RowType = row_type_from_string(
                    row[0].value
                )  # None if not a valid type

                # only pay attention if we have a reason to care
                if row_type == RowType.PREAMBLE:
                    preamble_rows.append(row[1:])

                elif row_type == RowType.HEADER:
                    # if it's already been defined, that's an error
                    if data_header is not None:
                        raise Exception(
                            f"Second header row encountered at #{n+1} in worksheet {config.name}"
                        )
                    else:
                        data_header = row[1:]

                elif row_type == RowType.DATA:
                    if data_header is None:
                        raise Exception(
                            f"Encountered data row (#{n+1} in worksheet {config.name!r}) before header row"
                        )
                    else:
                        data_rows.append(row[1:])

                # ignore everything else
                else:
                    pass

            # turn the preamble into a mapping
            preamble_values = {
                str(row[0].value).lower(): row[1].value
                for row in preamble_rows
                if row[0].value is not None
            }
            # check the shape of the preamble
            if len(preamble_values) != len(config.preamble):
                raise Exception(
                    f"Expected {len(config.preamble)} preamble rows but saw {len(preamble_values)} in worksheet {config.name}"
                )
            for entry in config.preamble:
                # process the value
                try:
                    new_value_map = entry.get_column_mapping(
                        preamble_values.get(entry.name.lower())
                    )

                    for col, val in new_value_map.items():
                        if isinstance(val, dict) and col in preamble_dict:
                            preamble_dict[col].update(val)
                        else:  # not isinstance(val, dict) or col not in preamble_dict
                            preamble_dict[col] = val

                except Exception as e:
                    # add a bit of context
                    raise Exception(
                        f"Error in processing preamble {entry.name} in worksheet {config.name}: {e.__cause__ if e.__cause__ is not None else e}"
                    )

            model_dicts.append(preamble_dict)

            # combine and flatten the lists of configs across all data_sections
            data_configs = [
                entry for entries in config.data_sections.values() for entry in entries
            ]
            if len(data_rows):
                header_width = len([c for c in data_header if c.value is not None])
                # check the shape of the data
                if header_width != len(data_configs):
                    raise Exception(
                        f"Expected {len(data_configs)} data columns but saw {header_width} in worksheet {config.name}"
                    )

            for n, row in enumerate(data_rows):
                data_values = {
                    str(title_cell.value).lower(): data_cell.value
                    for title_cell, data_cell in zip(data_header, row)
                    if title_cell.value is not None
                }

                # context will be updated by every cell for each row,
                # but preamble_dict needs to persist unchanged
                context = preamble_dict.copy()
                data_dict = {}
                for entry in data_configs:
                    # process the value
                    try:
                        new_value_map = entry.get_column_mapping(
                            data_values.get(entry.name.lower()), context
                        )

                        for col, val in new_value_map.items():
                            if isinstance(val, dict) and col in data_dict:
                                data_dict[col].update(val)
                            else:  # not isinstance(val, dict) or col not in data_dict
                                data_dict[col] = val
                    except Exception as e:
                        # add some context here
                        raise Exception(
                            f"Error in processing {entry.name} for data row #{n+1} in worksheet {config.name}: {e.__cause__ if e.__cause__ is not None else e}"
                        )
                    context.update(data_dict)

                model_dicts.append(data_dict)

            # now let's take all our data from this sheet and make some models out of them
            for model_dict in model_dicts:
                # separate by target class, regardless of class hierarchy
                model_groups = defaultdict(dict)
                for column, value in model_dict.items():
                    model_groups[column.class_][column.name] = value

                # kwargs is {column.name: value}
                for model, kwargs in model_groups.items():
                    # also grab all the kwargs for all superclasses
                    [
                        kwargs.update(other_kwargs)
                        for other_model, other_kwargs in model_groups.items()
                        if other_model in _all_bases(model)
                    ]
                    model_instances.append(model(**kwargs))

        # Group model instances with matching values in their primary key or
        # unique-constrained columns.
        model_groups: Dict[
            Type[MetadataModel], Dict[Tuple, List[MetadataModel]]
        ] = defaultdict(lambda: defaultdict(list))
        for instance in model_instances:
            unique_values: Tuple = instance.unique_field_values()
            model_groups[instance.__class__][unique_values].append(instance)

        # Build a dictionary mapping model classes to deduplicated instances
        # of that class. Make sure that all instances have valid pk's for insertion,
        # remembering that some fk's can't be set until insert time
        deduped_instances: Dict[Type[MetadataModel], List[MetadataModel]] = defaultdict(
            list
        )
        for model, groups in model_groups.items():
            # keep track of which columns can be set later by fk
            # can only set if there's only a single foreign instance
            # also look at all the superclasses for hidden fk's
            fk_to_check = [fk for fk in model.__table__.foreign_keys]
            fk_to_check.extend(
                [
                    fk
                    for b in _all_bases(model)
                    if hasattr(b, "__table__")
                    for fk in b.__table__.foreign_keys
                ]
            )
            columns_to_set_later_by_fk = []
            for k, v in deduped_instances.items():
                for fk in fk_to_check:
                    if len(v) == 1 and fk.column.table.name in [k.__tablename__] + [
                        b.__tablename__
                        for b in _all_bases(k)
                        if hasattr(b, "__tablename__")
                    ]:
                        columns_to_set_later_by_fk.append(fk.parent)

            # special value None for all(unique_values is None)
            broad_instances = groups.pop(None, [])
            for specific_instances in groups.values():
                instance = model()
                for partial_instance in broad_instances + specific_instances:
                    instance.merge(partial_instance)

                for pre_col, pre_val in preamble_dict.items():
                    if (
                        hasattr(instance, pre_col.name)
                        and getattr(instance, pre_col.name) is None
                    ):
                        setattr(instance, pre_col.name, pre_val)

                if all(
                    [
                        # need to validate ALL of the primary keys
                        any(
                            [
                                # ANY of these is a reason that the value is fine
                                pk is not None,  # if it's set
                                any(
                                    [
                                        c.name == fk_col.name
                                        for fk_col in columns_to_set_later_by_fk
                                    ]
                                ),  # if it'll be set later
                                c.server_default is not None,  # if it has a default
                                c.autoincrement is True,  # if it autoincrements
                            ]
                        )
                        for c, pk in instance.primary_key_map().items()
                    ]
                ):
                    deduped_instances[model].append(instance)

        # now order the models for insertion based on the calculated order
        ordered_instances: OrderedDict_Type[Type, List[MetadataModel]] = OrderedDict()
        for next_model in MODEL_INSERTION_ORDER:
            next_instances = deduped_instances.pop(next_model, None)
            if next_instances:
                ordered_instances[next_model] = next_instances

        return ordered_instances

    def write(self, filename: str):
        """
        Generate an empty excel file for this template type.
        """
        workbook = xlsxwriter.Workbook(filename)
        styles = ExcelStyles(workbook)
        self._write_legend(workbook, styles)
        enum_ranges = self._write_data_dict(workbook, styles)
        self._write_worksheets(workbook, styles, enum_ranges)
        workbook.close()

    def _write_legend(self, workbook: xlsxwriter.Workbook, styles: ExcelStyles):
        ws = workbook.add_worksheet("Legend")
        ws.protect()
        ws.set_column(1, 100, width=styles.COLUMN_WIDTH)

        row = 0
        ws.write(row, 1, f"LEGEND", styles.DATA_STYLE)

        for config in self.worksheet_configs:
            row += 1
            ws.write(row, 1, f"Legend for tab {config.name!r}", styles.TITLE_STYLE)

            for entry in config.preamble:
                row += 1
                self._write_legend_item(ws, row, entry, styles.PREAMBLE_STYLE)

            for section_name, section_entries in config.data_sections.items():
                row += 1
                ws.write(
                    row,
                    1,
                    f"Section {section_name!r} of tab {config.name!r}",
                    styles.DIRECTIVE_STYLE,
                )

                for entry in section_entries:
                    row += 1
                    self._write_legend_item(ws, row, entry, styles.HEADER_STYLE)

    def _write_legend_item(
        self,
        ws: xlsxwriter.workbook.Worksheet,
        row: int,
        entry: Entry,
        style: xlsxwriter.workbook.Format,
    ):
        """Writes a property with its type, description, and example if any."""
        ws.write(row, 1, entry.name, style)
        ws.write(row, 2, entry.sqltype.__class__.__name__)
        ws.write(row, 3, entry.doc)

    def _write_data_dict(
        self, workbook: xlsxwriter.Workbook, styles: ExcelStyles
    ) -> Dict[str, str]:
        ws = workbook.add_worksheet(self.DATA_DICT_SHEETNAME)
        ws.protect()
        ws.set_column(1, 100, width=styles.COLUMN_WIDTH)

        col = 0

        # a result dictionary that maps field names to data dictionary sheet
        # ranges of enum values to be used for validation
        data_dict_mapping = {}

        for config in self.worksheet_configs:
            for entry in config.preamble:
                rows = self._write_data_dict_item(ws, col, entry, styles.PREAMBLE_STYLE)
                if rows > 0:
                    # saving Data Dict range to use for validation
                    data_dict_mapping[entry.name] = _format_validation_range(
                        rows, col, self.DATA_DICT_SHEETNAME
                    )
                    col += 1

            for section_entries in config.data_sections.values():
                for entry in section_entries:
                    rows = self._write_data_dict_item(
                        ws, col, entry, styles.HEADER_STYLE
                    )
                    if rows > 0:
                        # saving Data Dict range to use for validation
                        data_dict_mapping[entry.name] = _format_validation_range(
                            rows, col, self.DATA_DICT_SHEETNAME
                        )
                        col += 1

        return data_dict_mapping

    def _write_data_dict_item(
        self,
        ws: xlsxwriter.workbook.Worksheet,
        col: int,
        entry: Entry,
        style: xlsxwriter.workbook.Format,
    ):
        if entry.enums is None:
            return 0

        # Write the data dict column header
        ws.write(0, col, entry.name, style)

        # Write the data dict column values
        for i, enum_value in enumerate(entry.enums):
            ws.write(1 + i, col, enum_value)

        return len(entry.enums)

    def _write_worksheets(
        self,
        workbook: xlsxwriter.Workbook,
        styles: ExcelStyles,
        enum_ranges: Dict[str, str],
    ):
        """Write content to the given worksheet"""
        for config in self.worksheet_configs:
            ws = workbook.add_worksheet(config.name)
            ws.set_column(0, 100, width=styles.COLUMN_WIDTH)

            row = 0
            col = 1

            # WORKSHEET TITLE
            self._write_type_annotation(ws, row, RowType.TITLE)
            preamble_range = xl_range(row, 1, row, 2)
            ws.merge_range(preamble_range, config.name, styles.TITLE_STYLE)

            # PREAMBLE ROWS
            row += 1
            for entry in config.preamble:
                # Write row type and entity name
                self._write_type_annotation(ws, row, RowType.PREAMBLE)
                ws.write(row, 1, entry.name, styles.PREAMBLE_STYLE)
                self._write_comment(ws, row, 1, entry, styles)

                # Format value cells next to entity name
                ws.write(row, 2, "", styles.PREAMBLE_STYLE)

                # Add data validation if appropriate
                value_cell = xl_rowcol_to_cell(row, 2)
                self._write_validation(ws, value_cell, entry, enum_ranges)
                row += 1

            # DATA SECTION HEADERS
            row += 1
            self._write_type_annotation(ws, row, RowType.SKIP)
            start_col = 1
            for section_header, section_entries in config.data_sections.items():
                section_width = len(section_entries)
                end_col = start_col + section_width - 1
                if end_col - start_col > 0:
                    ws.merge_range(
                        row,
                        start_col,
                        row,
                        end_col,
                        section_header,
                        styles.DIRECTIVE_STYLE,
                    )
                else:
                    ws.write(row, start_col, section_header, styles.DIRECTIVE_STYLE)
                start_col = end_col + 1

            # DATA SECTION TYPE ANNOTATIONS
            row += 1
            self._write_type_annotation(ws, row, RowType.HEADER)
            annotations = [RowType.DATA.value] * self.DATA_ROWS
            ws.write_column(row + 1, 0, annotations)

            # DATA SECTION ROWS
            for section_entries in config.data_sections.values():
                for entry in section_entries:
                    ws.write(row, col, entry.name, styles.HEADER_STYLE)
                    self._write_comment(ws, row, col, entry, styles)

                    # Write validation to data cells below header cell
                    data_range = xl_range(row + 1, col, row + self.DATA_ROWS, col)
                    self._write_validation(ws, data_range, entry, enum_ranges)
                    col += 1

    def _write_type_annotation(
        self, ws: xlsxwriter.workbook.Worksheet, row: int, row_type: RowType
    ):
        ws.write(row, 0, row_type.value)

    def _write_comment(
        self,
        ws: xlsxwriter.workbook.Worksheet,
        row: int,
        col: int,
        entry: Entry,
        styles: ExcelStyles,
    ):
        comment = entry.doc

        if entry.gcs_uri_format is not None:
            if isinstance(entry.gcs_uri_format, str):
                comment += f'\nIn .{entry.gcs_uri_format.split(".")[-1]} format'
            elif isinstance(entry.gcs_uri_format, dict):
                if "template_comment" in entry.gcs_uri_format:
                    comment += "\n" + entry.gcs_uri_format["template_comment"]

        if comment:
            ws.write_comment(row, col, comment, styles.COMMENT_STYLE_PROPS)

    def _write_validation(
        self,
        ws: xlsxwriter.workbook.Worksheet,
        cell_range: str,
        entry: Entry,
        enum_ranges: Dict[str, str],
    ):
        validation = self._get_validation(cell_range, entry, enum_ranges)
        if validation:
            ws.data_validation(cell_range, validation)

    def _get_validation(
        self, cell_range: str, entry: Entry, data_dict_validations: dict
    ) -> Optional[dict]:
        if entry.enums and len(entry.enums) > 0:
            data_dict_validation_range = data_dict_validations[entry.name]
            return {"validate": "list", "source": data_dict_validation_range}

        elif entry.pytype == datetime.date:
            return {
                "validate": "custom",
                "value": self._make_date_validation_string(cell_range),
                "error_message": "Please enter date in format mm/dd/yyyy",
            }
        elif entry.pytype == datetime.time:
            return {
                "validate": "time",
                "criteria": "between",
                "minimum": datetime.time(0, 0),
                "maximum": datetime.time(23, 59),
                "error_message": "Please enter time in format hh:mm",
            }
        elif entry.pytype == bool:
            return {"validate": "list", "source": ["True", "False"]}

        return None

    @staticmethod
    def _make_date_validation_string(cell_range: str) -> str:
        return f'=AND(ISNUMBER({cell_range}),LEFT(CELL("format",{cell_range}),1)="D")'



__all__ = [
    "_all_bases",
    "_all_subclasses",
    "_get_global_insertion_order",
    "in_single_transaction",
    "insert_record_batch",
    "get_full_template_name",
    "remove_record_batch",
]

from collections import defaultdict
from typing import Any, Callable, Dict, List, OrderedDict, Set, Type

from psycopg2.errors import (
    CheckViolation,
    ForeignKeyViolation,
    InvalidTextRepresentation,
    NotNullViolation,
)
from sqlalchemy.exc import DataError, IntegrityError
from sqlalchemy.orm import Session

from .file_metadata import Upload
from .model_core import MetadataModel, with_default_session


def _all_bases(cls: Type) -> Set[Type]:
    return set(cls.__bases__).union([s for c in cls.__bases__ for s in _all_bases(c)])


def _all_subclasses(cls: Type) -> Set[Type]:
    return set(cls.__subclasses__()).union(
        [s for c in cls.__subclasses__() for s in _all_subclasses(c)]
    )


def _get_global_insertion_order() -> List[MetadataModel]:
    """
    Produce an ordering of all metadata model types based on foreign key dependencies
    between models. For a given model, all models it depends on are guaranteed to
    appear before it in this list.
    """
    models = _all_subclasses(MetadataModel)

    # Build a dictionary mapping table names to model classes -
    # we use this below to look up the model classes associated with
    # a given foreign key.
    table_to_model = defaultdict(list)
    for m in models:
        table_to_model[m.__tablename__].append(m)

    # Build a graph representing foreign-key relationships,
    # mapping models to the set of models they depend on
    fks_on_model = defaultdict(set)
    fks_to_parents = defaultdict(set)
    for model in models:
        fks_on_model[model] = model.__table__.foreign_keys.union(
            {
                fk
                for b in _all_bases(model)
                if hasattr(b, "__table__")
                for fk in b.__table__.foreign_keys
            }
        )
        for fk in fks_on_model[model]:
            fk_models = table_to_model.get(fk.column.table.name, [])
            for fk_model in fk_models:
                if not issubclass(fk_model, model):
                    fks_to_parents[model].add(fk_model)

                # special handling to make sure all Upload subclasses get merged earlier
                # class inheritance will take care of relations within Upload subclasses
                ## created to prevent Upload from being created after relevant Files that need upload_id
                if issubclass(model, Upload) and not issubclass(fk_model, Upload):
                    for c in _all_subclasses(Upload):
                        if c is not fk_model and not issubclass(fk_model, c):
                            fks_to_parents[c].add(fk_model)

                if issubclass(fk_model, Upload) and not issubclass(model, Upload):
                    for c in _all_subclasses(Upload):
                        if c is not model and not issubclass(c, model):
                            fks_to_parents[model].add(c)

    # Topologically sort the dependency graph to produce a valid insertion order
    # using Kahn's algorithm: https://en.wikipedia.org/wiki/Topological_sorting#Kahn's_algorithm
    ordered_models = []
    depless_models = set(model for model in models if len(fks_to_parents[model]) == 0)
    while len(depless_models) > 0:
        adding = depless_models.pop()
        ordered_models.append(adding)

        for model, fk_models in fks_to_parents.items():
            if adding in fk_models:
                fks_to_parents[model].remove(adding)

                if len(fks_to_parents[model]) == 0:
                    depless_models.add(model)

        fks_to_parents.pop(adding)

    if len(fks_to_parents) > 0:
        raise Exception(f"Cannot figure out how to insert: {fks_to_parents}")

    return ordered_models


def _handle_postgres_error(error: IntegrityError, model: Type) -> Exception:
    if not hasattr(error, "orig"):
        return error

    orig = error.orig
    pg_error: str = orig.pgerror
    instance = {c.name: v for c, v in model(**error.params).primary_key_map().items()}

    if isinstance(orig, ForeignKeyViolation):
        # ERROR:  insert or update on table "<table>" violates foreign key constraint "<name>_fkey"\n
        # DETAIL: Key (<column>)=(<value>) is not present in table "<foreign>".\n
        foreign_name = pg_error.split('"')[-2].replace("_", " ").title()[:-1]
        _, column_name, _, value, _ = pg_error.replace(")", "(").split("(")
        instance = {
            c.name: v for c, v in model(**error.params).primary_key_map().items()
        }
        return Exception(
            f"Error inserting / updating {model.__name__}: no {foreign_name} with {column_name} = {value}\n{instance}"
        )

    elif isinstance(orig, NotNullViolation):
        # ERROR: null value in column "<column>" violates not-null constraint
        # DETAIL: Failing row contains (<value>, ...).
        column_name = pg_error.split('"')[1]
        return Exception(
            f"Error inserting / updating {model.__name__}: must provide a value for {column_name}\n{instance}"
        )

    elif isinstance(orig, InvalidTextRepresentation):
        # ERROR: invalid input value for enum <whatever>_enum: "<value>"
        # DETAIL: LINE 1: ..., '<value>', ...
        #                      ^
        value = pg_error.split('"')[1]
        column_name = {str(v): k for k, v in error.params.items()}[value]
        return Exception(
            f"Error inserting / updating {model.__name__}: invalid input value for {column_name} = {value}\n{instance}"
        )

    elif isinstance(orig, CheckViolation):
        ## currently have to be manually added to the db itself as sqlalchemy doesn't handle these
        # ERROR: new row for relation "<table>" violates check constraint "<name>"
        # DETAIL: Failing row contains (<value>, ...).
        constraint_name = pg_error.split('"')[3]
        return Exception(
            f"Error inserting / updating {model.__name__}: improper value checked by {constraint_name}\n{instance}"
        )

    else:
        return error


@with_default_session
def insert_record_batch(
    ordered_records: OrderedDict[Type, List[MetadataModel]],
    *,
    dry_run: bool = False,
    hold_commit: bool = False,
    session: Session,
) -> List[Exception]:
    """
    Try to insert the given list of models into the database in a single transaction,
    rolling back and returning a list of errors if any are encountered. If `dry_run` is `True`,
    rollback the transaction regardless of whether any errors are encountered.
    Uses session.merge, so not a "true" insert.
    If `hold_commit` is passed, all rollback / commit are ignored.
    """
    errors = []
    for model in ordered_records.keys():
        records = ordered_records[model]

        # set any columns that were left to fill in by fk
        # can only set if there's only a single foreign instance
        # also look at all the superclasses for hidden fk's
        fk_to_check = [fk for fk in model.__table__.foreign_keys]
        fk_to_check.extend(
            [
                fk
                for b in _all_bases(model)
                if hasattr(b, "__table__")
                for fk in b.__table__.foreign_keys
            ]
        )
        for fk, target_class in {
            fk: k
            for k, v in ordered_records.items()
            for fk in fk_to_check
            if len(v) == 1
            and fk.column.table.name
            in [k.__tablename__]
            + [b.__tablename__ for b in _all_bases(k) if hasattr(b, "__tablename__")]
        }.items():
            for n in range(len(records)):
                setattr(
                    records[n],
                    fk.parent.name,
                    getattr(ordered_records[target_class][0], fk.column.name),
                )

        # merge all records into session and keep a copy
        for n, record in enumerate(records):
            try:
                record = session.merge(record)
                ordered_records[model][n] = record
            except Exception as e:
                errors.append(_handle_postgres_error(e, model))
                break

        if len(errors) != 0:  # if we've hit an error, we're done
            break
        else:
            # flush these records to generate db-derived values
            # in case they're needed for later fk's
            try:
                session.flush()
            except (DataError, IntegrityError) as e:
                errors.append(_handle_postgres_error(e, model=model))
                break  # if it fails in a flush, it's done done

    if len(errors):
        session.rollback()
    elif hold_commit:
        try:
            session.flush()
        except Exception as e:
            errors.append(_handle_postgres_error(e, model))
            session.rollback()

    elif dry_run:
        try:
            session.flush()
        except Exception as e:
            errors.append(_handle_postgres_error(e, model))
        session.rollback()

    else:
        try:
            session.commit()
        except Exception as e:
            errors.append(_handle_postgres_error(e, model))
            session.rollback()

    return errors


@with_default_session
def remove_record_batch(
    records: List[MetadataModel],
    *,
    dry_run: bool = False,
    hold_commit: bool = False,
    session: Session,
) -> List[Exception]:
    """
    Try to safely remove the given list of models from the database in a single transaction,
    rolling back and returning a list of errors if any are encountered. If `dry_run` is `True`,
    rollback the transaction regardless of whether any errors are encountered.
    If `hold_commit` is passed, all rollback / commit are ignored.
    """
    errors = []

    # merge all records into session and keep a copy
    for record in records:
        if record is None:
            continue

        try:
            record = session.delete(record)
        except Exception as e:
            errors.append(e)

    try:
        if hold_commit:
            session.flush()
        elif dry_run or len(errors):
            session.flush()
            session.rollback()
        else:
            session.commit()
    except Exception as e:
        errors.append(e)
        session.rollback()

    return errors


@with_default_session
def in_single_transaction(
    calls: OrderedDict[Callable[[Any], List[Exception]], Dict[str, Any]],
    *,
    dry_run: bool = False,
    session: Session,
) -> List[Exception]:
    """Given an arbitrary set of calls, make all of them in a single transaction,
    rolling back and returning a list of errors if any are encountered. If `dry_run` is `True`,
    rollback the transaction regardless of whether any errors are encountered.
    """
    errors = []
    for func, kwargs in calls.items():
        kwargs.update({"session": session, "hold_commit": True})
        errors.extend(func(**kwargs))

        try:
            session.flush()
        except Exception as e:
            errors.append(e)
            break

    # no hold_commit here because we need to close the transaction
    if dry_run or len(errors):
        session.rollback()
    else:
        try:
            session.commit()
        except Exception as e:
            errors.append(e)
            session.rollback()

    return errors


def get_full_template_name(temp_name: str):
    """Returns a fully qualified name of a metadata template in the form '<assay_name>'
    where <assay_name> is snake_case
    """
    # as AbbrevCamelCasePurpose
    for purpose in ["Assay", "Analysis", "Manifest"]:
        if temp_name.endswith(purpose):
            temp_name = temp_name[: -len(purpose)]
            break

    return "".join(
        [
            "_" + i if i != j and n else i
            for n, (i, j) in enumerate(zip(temp_name, temp_name.lower()))
        ]
    ).lower()  # as abbrev_camel_case



__all__ = [
    "Aliquot",
    "AssaysEnum",
    "ClinicalTrial",
    "Cohort",
    "CollectionEvent",
    "ConcentrationUnits",
    "MaterialUnits",
    "Participant",
    "Replacement",
    "Sample",
    "Shipment",
    "VolumeUnits",
]

import re
from sqlalchemy import (
    CheckConstraint,
    Column,
    Date,
    Enum,
    ForeignKey,
    ForeignKeyConstraint,
    Integer,
    Numeric,
    String,
    UniqueConstraint,
)
from sqlalchemy.orm import relationship
from sqlalchemy.orm.session import Session
from sqlalchemy.dialects.postgresql import JSONB
from typing import List

from .model_core import MetadataModel, with_default_session

cimac_id_regex_pattern = "^C[A-Z0-9]{3}[A-Z0-9]{3}[A-Z0-9]{2}.[0-9]{2}$"
cimac_id_regex = re.compile(cimac_id_regex_pattern)

AssaysEnum = Enum(
    "ATACseq",
    "CyTOF",
    "ELISA",
    "H&E",
    "IHC",
    "mIF",
    "mIHC",
    "Olink",
    "RNAseq",
    "TCRseq",
    "WES",
    name="assay_enum",
)
ConcentrationUnits = Enum(
    "Nanogram per Microliter",
    "Milligram per Milliliter",
    "Micrograms per Microliter",
    "Cells per Vial",
    "Not Reported",
    "Other",
    name="concentration_units_enum",
)
MaterialUnits = Enum(
    "Microliters",
    "Milliliters",
    "Nanogram per Microliter",
    "Milligram per Milliliter",
    "Micrograms per Microliter",
    "Cells per Vial",
    "Slides",
    "Not Reported",
    "Other",
    name="material_units_enum",
)
Replacement = Enum(
    "Replacement Not Requested",
    "Replacement Requested",
    "Replacement Tested",
    "Not Reported",
    "Other",
    name="replace_enum",
)
VolumeUnits = Enum(
    "Microliters", "Milliliters", "Not Reported", "Other", name="volume_units_enum"
)


class ClinicalTrial(MetadataModel):
    """A clinical trial in the CIMAC-CIDC network."""

    __tablename__ = "clinical_trials"

    protocol_identifier = Column(
        String,
        primary_key=True,  # allows for use as Foreign Key
        doc="Trial identifier used by lead organization, ie. Center for Experimental Therapeutics Program (CTEP) ID or Industry Sponsored ID.  This is usually a short identifier. Example: E4412.",
    )
    nct_id = Column(String, doc="ClinicalTrials.gov identifier. Example: NCT03731260.")
    nci_id = Column(String, doc="NCI Trial Identifier. Example: NCI22345.")
    trial_name = Column(String, doc="Name of clinical trial.")
    trial_description = Column(String, doc="A brief description of the clinical trial.")
    trial_organization = Column(
        String,
        doc="Name of the primary organization that oversees the clinical trial. Example: ECOG-ACRIN, SWOG, etc.",
    )
    grant_or_affiliated_network = Column(
        String,
        doc="The primary organization providing grant funding and supporting the trial.",
    )
    trial_status = Column(
        String, doc="What stage the clinical trial is at in its process."
    )
    biobank = Column(
        String,
        doc="The primary organization responsible for storing biospecimens from this study.",
    )
    lead_cimac_pis = Column(
        String, doc="The PI(s) from the CIMAC-CIDC network responsible for this study."
    )
    lead_cimac_contacts = Column(String, doc="A list of contacts for this trial.")
    lead_trial_staff = Column(
        String, doc="The names of lead trial staff members other than the PIs."
    )
    justification = Column(
        String,
        doc="A description of the reasons why this study could provide insight into molecular biomarkers of immunotherapy.",
    )
    biomarker_plan = Column(
        String,
        doc="A description of the objectives and hypotheses for the proposed biomarkers.",
    )
    data_sharing_plan = Column(
        String,
        doc="A description of the rules governing data sharing and publications.",
    )

    # expected_assays : List[str], doc="A list of assays the CIDC expects to receive for this trial."
    # assays : List[Assay], doc="Assays for this trial"
    # analysis : List[Analysis], doc="Analyses for this trial"
    # clinical_data : ClinicalData, doc="Clinical data for this trial"
    # schema : artifact_image, doc="An image of the schema of this trial."

    cohorts = relationship(
        "Cohort",
        back_populates="trial",
        sync_backref=False,
        viewonly=True,
        doc="The collection of all cohorts related to this trial.",
    )
    collection_events = relationship(
        "CollectionEvent",
        back_populates="trial",
        sync_backref=False,
        viewonly=True,
        doc="The collection of all collection events related to this trial.",
    )
    shipments = relationship(
        "Shipment",
        back_populates="trial",
        sync_backref=False,
        viewonly=True,
        doc="The collection of all shipments related to this trial.",
    )
    participants = relationship(
        "Participant",
        back_populates="trial",
        sync_backref=False,
        viewonly=True,
        doc="The collection of participants in this trial.",
    )

    @property
    @with_default_session
    def allowed_cohort_names(self, *, session: Session) -> List[str]:
        """Allowed values for Participant.cohort_name for this trial."""
        return [
            c.cohort_name
            for c in session.query(Cohort)
            .filter(Cohort.trial_id == self.protocol_identifier)
            .all()
        ]

    @property
    @with_default_session
    def allowed_collection_event_names(self, *, session: Session) -> List[str]:
        """Allowed values for Sample.collection_event_name for this trial."""
        return [
            ce.event_name
            for ce in session.query(CollectionEvent)
            .filter(CollectionEvent.trial_id == self.protocol_identifier)
            .all()
        ]


class Cohort(MetadataModel):
    __tablename__ = "cohorts"

    trial_id = Column(
        String, ForeignKey(ClinicalTrial.protocol_identifier), primary_key=True
    )
    cohort_name = Column(
        String, primary_key=True
    )  # both True allows for use as multi Foreign Key

    trial = relationship(
        ClinicalTrial, back_populates="cohorts", sync_backref=False, viewonly=True
    )
    participants = relationship(
        "Participant", back_populates="cohort", sync_backref=False, viewonly=True
    )


class CollectionEvent(MetadataModel):
    __tablename__ = "collection_events"

    trial_id = Column(
        String, ForeignKey(ClinicalTrial.protocol_identifier), primary_key=True
    )
    event_name = Column(
        String, primary_key=True
    )  # both True allows for use as multi Foreign Key

    samples = relationship(
        "Sample", back_populates="collection_event", sync_backref=False, viewonly=True
    )
    trial = relationship(
        ClinicalTrial,
        back_populates="collection_events",
        sync_backref=False,
        viewonly=True,
    )


class Shipment(MetadataModel):
    __tablename__ = "shipments"

    trial_id = Column(
        String, ForeignKey(ClinicalTrial.protocol_identifier), primary_key=True
    )
    manifest_id = Column(
        String,
        primary_key=True,  # both True allows for use as multi Foreign Key
        unique=True,
        doc="Filename of the manifest used to ship this sample. Example: E4412_PBMC.",
    )
    json_data = Column(JSONB, nullable=False, default={})

    __table_args__ = (
        UniqueConstraint(trial_id, manifest_id, name="unique_trial_manifest"),
    )

    assay_priority = Column(
        Enum(
            "1",
            "2",
            "3",
            "4",
            "5",
            "6",
            "7",
            "8",
            "9",
            "10",
            "11",
            "12",
            "13",
            "14",
            "15",
            "Not Reported",
            "Other",
            name="assay_priority_enum",
        ),
        doc="Priority of the assay as it appears on the intake form.",
    )
    assay_type = Column(AssaysEnum, nullable=True, doc="Assay and sample type used.")
    courier = Column(
        Enum("FEDEX", "USPS", "UPS", "Inter-Site Delivery", name="courier_enum"),
        doc="Courier utilized for shipment.",
    )
    tracking_number = Column(
        String,
        doc="Air bill number assigned to shipment. Example: 4567788343.",
    )
    account_number = Column(
        String,
        doc="Courier account number to pay for shipping if available. Example: 45465732.",
    )
    shipping_condition = Column(
        Enum(
            "Frozen_Dry_Ice",
            "Frozen_Shipper",
            "Ice_Pack",
            "Ambient",
            "Not Reported",
            "Other",
            name="shipping_condition_enum",
        ),
        doc="Type of shipment made.",
    )
    date_shipped = Column(Date, doc="Date of shipment.")
    date_received = Column(Date, doc="Date of receipt.")
    quality_of_shipment = Column(
        Enum(
            "Specimen shipment received in good condition",
            "Specimen shipment received in poor condition",
            "Not Reported",
            "Other",
            name="quality_of_shipment_enum",
        ),
        doc="Indication that specimens were received in good condition.",
    )
    ship_from = Column(String, doc="Contact information for shipment.")
    ship_to = Column(String, doc="Physical shipping address of the destination.")
    receiving_party = Column(
        Enum(
            "MDA_Wistuba",
            "MDA_Bernatchez",
            "MDA_Al-Atrash",
            "MSSM_Gnjatic",
            "MSSM_Rahman",
            "MSSM_Kim-Schulze",
            "MSSM_Bongers",
            "DFCI_Wu",
            "DFCI_Hodi",
            "DFCI_Severgnini",
            "DFCI_Livak",
            "Broad_Cibulskis",
            "Stanf_Maecker",
            "Stanf_Bendall",
            "NCH",
            "Adaptive",
            "FNLCR_MoCha",
            name="receiving_party_enum",
        ),
        doc="Site where sample was shipped to be assayed.",
    )

    trial = relationship(
        ClinicalTrial, back_populates="shipments", sync_backref=False, viewonly=True
    )
    samples = relationship(
        "Sample", back_populates="shipment", sync_backref=False, viewonly=True
    )


class Participant(MetadataModel):
    __tablename__ = "participants"

    trial_id = Column(
        String, ForeignKey(ClinicalTrial.protocol_identifier), primary_key=True
    )
    cimac_participant_id = Column(
        String,
        CheckConstraint("cimac_participant_id ~ '^C[A-Z0-9]{3}[A-Z0-9]{3}$'"),
        primary_key=True,  # both True allows for use as multi Foreign Key
        doc="Participant identifier assigned by the CIMAC-CIDC Network. Formated as: C?????? (first 7 characters of CIMAC ID)",
    )
    trial_participant_id = Column(
        String,
        nullable=True,
        doc="Trial Participant Identifier. Crypto-hashed after upload.",
    )
    cohort_name = Column(String, nullable=True)
    json_data = Column(JSONB, nullable=False, default={})

    __table_args__ = (
        UniqueConstraint(
            trial_id, cimac_participant_id, name="unique_trial_participant"
        ),
        ForeignKeyConstraint(
            [trial_id, cohort_name], [Cohort.trial_id, Cohort.cohort_name]
        ),
    )

    gender = Column(
        Enum("Male", "Female", "Not Specified", "Other", name="gender_enum"),
        doc="Identifies the gender of the participant.",
    )
    race = Column(
        Enum(
            "American Indian/Alaska Native",
            "Asian",
            "Black/African American",
            "Native Hawaiian/Pacific Islander",
            "White",
            "Not Reported",
            "Unknown",
            "Other",
            name="race_enum",
        ),
        doc="NIH Racial and Ethnic Categories and Definitions for NIH Diversity Programs and for Other Reporting Purposes (NOT-OD-15-089),  Release Date: April 8, 2015.",
    )
    ethnicity = Column(
        Enum(
            "Hispanic or Latino",
            "Not Hispanic or Latino",
            "Not reported",
            "Unknown",
            "Other",
            name="ethnicity_enum",
        ),
        doc="NIH Racial and Ethnic Categories and Definitions for NIH Diversity Programs and for Other Reporting Purposes (NOT-OD-15-089),  Release Date: April 8, 2015.",
    )

    cohort = relationship(
        Cohort, back_populates="participants", sync_backref=False, viewonly=True
    )
    samples = relationship(
        "Sample", back_populates="participant", sync_backref=False, viewonly=True
    )
    trial = relationship(
        ClinicalTrial, back_populates="participants", sync_backref=False, viewonly=True
    )


class Sample(MetadataModel):
    __tablename__ = "samples"

    trial_id = Column(String, primary_key=True)
    cimac_id = Column(
        String,
        primary_key=True,  # both True allows for use as multi Foreign Key
        unique=True,
        doc="Specimen identifier assigned by the CIMAC-CIDC Network. Formatted as C????????.??",
    )
    cimac_participant_id = Column(String, nullable=False)
    manifest_id = Column(String, nullable=False)
    collection_event_name = Column(String, nullable=True)
    json_data = Column(JSONB, nullable=False, default={})

    __table_args__ = (
        ForeignKeyConstraint(
            [trial_id, cimac_participant_id],
            [Participant.trial_id, Participant.cimac_participant_id],
        ),
        ForeignKeyConstraint(
            [trial_id, collection_event_name],
            [CollectionEvent.trial_id, CollectionEvent.event_name],
        ),
        ForeignKeyConstraint(
            [trial_id, manifest_id], [Shipment.trial_id, Shipment.manifest_id]
        ),
        CheckConstraint("cimac_id ~ '^C[A-Z0-9]{3}[A-Z0-9]{3}[A-Z0-9]{2}.[0-9]{2}$'"),
    )

    shipping_entry_number = Column(
        Integer,
        doc="Provides a numbered identifier for patient (sample) entry in a shipment manifest.",
    )
    box_number = Column(
        String,
        doc="Identifier if sample shipment container includes multiple boxes for each assay.",
    )
    surgical_pathology_report_id = Column(
        String,
        doc="A unique identifier so someone can find the surgical pathology report.",
    )
    clinical_report_id = Column(
        String, doc="A unique identifier so someone can find the clinical report."
    )
    parent_sample_id = Column(
        String,
        doc="Sample identifier assigned by the biorepository site. Crypto-hashed after upload.",
    )
    processed_sample_id = Column(
        String,
        doc="Aliquot identifier assigned by the biorepository site. Crypto-hashed after upload.",
    )
    site_description = Column(
        String, doc="Descritpion of the topography category. e.g LUNG AND BRONCHUS"
    )
    topography_code = Column(
        String,
        doc="ICD-0-3 topography site code from which a specimen was isolated. e.g. C34.1",
    )
    topography_description = Column(
        String, doc="ICD-0-3 site code description. e.g. Upper lobe, lung"
    )
    histology_behavior = Column(
        String, doc="ICD-0-3 code for histology and behavior. e.g. 9665/3"
    )
    histology_behavior_description = Column(
        String,
        doc="ICD-0-3 histology and behavior code description. e.g. Hodgkin lymphoma, nod. scler., grade 1",
    )
    sample_location = Column(
        String,
        doc="Sample location within the shipping container. Example: A1.",
    )
    type_of_sample = Column(
        Enum(
            "Tumor Tissue",
            "Normal Tissue",
            "Skin Tissue",
            "Blood",
            "Bone Marrow",
            "Cerebrospinal Fluid",
            "Lymph Node",
            "Stool",
            "Cell Product",
            "White Blood Cell Apheresis",
            "Not Reported",
            "Other",
            name="sample_types_enum",
        ),
        nullable=True,
        doc="Type of sample sent.",
    )
    type_of_tumor_sample = Column(
        Enum(
            "Metastatic Tumor",
            "Primary Tumor",
            "Not Reported",
            "Other",
            name="type_of_tumor_sample_enum",
        ),
        doc="The type of tumor sample obtained (primary or metastatic).",
    )
    sample_collection_procedure = Column(
        Enum(
            "Blood Draw",
            "Excision",
            "Core Biopsy",
            "Punch Biopsy",
            "Endoscopic Biopsy",
            "Bone Marrow Core Biopsy",
            "Bone Marrow Aspirate",
            "Lumbar Puncture",
            "Aspirate",
            "Fine-Needle Aspiration",
            "Not Reported",
            "Other",
            name="sample_collection_procedure_enum",
        ),
        doc="Indicates the specimen source of the sample shipped. Example: Na Heparin blood draw aliquots (2 of three), FFPE block #52",
    )
    core_number = Column(
        Integer, doc="The biopsy core number from which the sample was used."
    )
    fixation_stabilization_type = Column(
        Enum(
            "Archival FFPE",
            "Fresh Specimen",
            "Frozen Specimen",
            "Formalin-Fixed Paraffin-Embedded",
            "Optimum cutting temperature medium",
            "Thaw-Lyse",
            "Not Reported",
            "Other",
            name="fixation_stabilization_type_enum",
        ),
        doc="Type of specimen fixation or stabilization that was employed by the site directly after collection.",
    )
    type_of_primary_container = Column(
        Enum(
            "Sodium heparin",
            "Blood specimen container with EDTA",
            "Potassium EDTA",
            "Streck Blood Collection Tube",
            "Stool collection container with DNA stabilizer",
            "Not Reported",
            "Other",
            name="type_of_primary_container_enum",
        ),
        CheckConstraint(
            "type_of_sample != 'blood' or type_of_primary_container is not null"
        ),
        doc="The format in which the sample was sent.",
    )
    sample_volume = Column(
        Numeric, doc="Volume of the parent sample (e.g. Heparin tube volume)"
    )
    sample_volume_units = Column(VolumeUnits, doc="Unit for the parent sample volume.")
    processed_sample_type = Column(
        Enum(
            "Whole Blood",
            "Plasma",
            "PBMC",
            "Buffy Coat",
            "Bone Marrow Mononuclear Cell",
            "Supernatant",
            "Cell Pellet",
            "H&E-Stained Fixed Tissue Slide Specimen",
            "Fixed Slide",
            "Tissue Scroll",
            "FFPE Punch",
            "Not Reported",
            "Other",
            name="processed_sample_type_enum",
        ),
        doc="The type of processing that was performed on the collected specimen by the Biobank for storage.",
    )
    processed_sample_volume = Column(Numeric, doc="Volume of the processed sample.")
    processed_sample_volume_units = Column(
        VolumeUnits, doc="Volume units of the processed sample."
    )
    processed_sample_concentration = Column(
        Numeric, doc="The concentration of the processed sample."
    )
    processed_sample_concentration_units = Column(
        ConcentrationUnits, doc="The concentration units for the processed sample."
    )
    processed_sample_quantity = Column(
        Numeric,
        doc="Quantity of the processed sample (e.g. number of slides cut for DNA extraction).",
    )
    processed_sample_derivative = Column(
        Enum(
            "Tumor DNA",
            "Tumor RNA",
            "Germline DNA",
            "Circulating Tumor-Derived DNA",
            "Not Reported",
            "Other",
            name="processed_sample_derivative_enum",
        ),
        doc="The type of derivative or analyte extracted from the specimen to be shipped for testing.",
    )
    sample_derivative_volume = Column(
        Numeric, doc="Volume of the analyte or derivative shipped."
    )
    sample_derivative_volume_units = Column(
        VolumeUnits, doc="Volume units of each analyte or derivative shipped."
    )
    sample_derivative_concentration = Column(
        Numeric, doc="The concentration of analyte or derivative shipped."
    )
    sample_derivative_concentration_units = Column(
        ConcentrationUnits,
        doc="The concentration units for the analyte or derivative shipped.",
    )
    tumor_tissue_total_area_percentage = Column(
        Numeric,
        CheckConstraint(
            "tumor_tissue_total_area_percentage >= 0 and tumor_tissue_total_area_percentage <= 100"
        ),
        doc="Score the percentage of tumor (including tumor bed) tissue area of the slide (e.g. vs non-malignant or normal tissue)",
    )
    viable_tumor_area_percentage = Column(
        Numeric,
        CheckConstraint(
            "viable_tumor_area_percentage >= 0 and viable_tumor_area_percentage <= 100"
        ),
        doc="Score the percentage of viable tumor cells comprising the tumor bed area",
    )
    viable_stroma_area_percentage = Column(
        Numeric,
        CheckConstraint(
            "viable_stroma_area_percentage >= 0 and viable_stroma_area_percentage <= 100"
        ),
        doc="Score the evaluation of stromal elements (this indicates the % area of tumor bed occupied by non-tumor cells, including inflammatory cells [lymphocytes, histiocytes, etc], endothelial cells, fibroblasts, etc)",
    )
    necrosis_area_percentage = Column(
        Numeric,
        CheckConstraint(
            "necrosis_area_percentage >= 0 and necrosis_area_percentage <= 100"
        ),
        doc="Score the percentage area of necrosis",
    )
    fibrosis_area_percentage = Column(
        Numeric,
        CheckConstraint(
            "fibrosis_area_percentage >= 0 and fibrosis_area_percentage <= 100"
        ),
        doc="Score the percentage area of Fibrosis",
    )
    din = Column(
        Numeric,
        CheckConstraint("din >= 0 and din <= 10"),
        doc="Provides a DNA Integrity Numeric as an indication of extraction quality (values of 1-10)",
    )
    a260_a280 = Column(
        Numeric,
        CheckConstraint("a260_a280 >= 0 and a260_a280 <= 2"),
        doc="Provides an absorbance percentage ratio indicating purity of DNA (values of 0 to 2)",
    )
    a260_a230 = Column(
        Numeric,
        CheckConstraint("a260_a230 >= 0 and a260_a230 <= 3"),
        doc="Provides an absorbance percentage ratio indicating presence of contaminants (values of 0 to 3)",
    )
    pbmc_viability = Column(
        Numeric,
        CheckConstraint("pbmc_viability >= 0 and pbmc_viability <= 100"),
        doc="Receiving site determines the percent recovered cells that are viable after thawing.",
    )
    pbmc_recovery = Column(
        Numeric,
        doc="Receiving site determines number for PBMCs per vial recovered upon receipt.",
    )
    pbmc_resting_period_used = Column(
        Enum(
            "Yes", "No", "Not Reported", "Other", name="pbmc_resting_period_used_enum"
        ),  # should be Boolean, nullable=True
        doc="Receiving site indicates if a resting period was used after PBMC recovery.",
    )
    material_used = Column(
        Numeric,
        doc="Receiving site indicates how much material was used for assay purposes.",
    )
    material_used_units = Column(
        MaterialUnits,
        doc="Units for the amount of material used; should be the same value as Specimen Analyte units.",
    )
    material_remaining = Column(
        Numeric,
        doc="Receiving site indicates how much material remains after assay use.",
    )
    material_remaining_units = Column(
        MaterialUnits, doc="Units for the amount of material remaining."
    )
    material_storage_condition = Column(
        Enum(
            "RT",
            "4oC",
            "(-20)oC",
            "(-80)oC",
            "LN",
            "Not Reported",
            "Other",
            name="material_storage_condition_enum",
        ),
        doc="Storage condition of the material once it was received.",
    )
    quality_of_sample = Column(
        Enum(
            "Pass", "Fail", "Not Reported", "Other", name="quality_of_sample_enum"
        ),  # could be Boolean, nullable=True
        doc="Final status of sample after QC and pathology review.",
    )
    sample_replacement = Column(
        Replacement, doc="Indication if sample replacement is/was requested."
    )
    residual_sample_use = Column(
        Enum(
            "Sample Returned",
            "Sample Sent to Another Lab",
            "Sample received from CIMAC",
            "Not Reported",
            "Other",
            name="residual_sample_use_enum",
        ),
        doc="Indication if sample was sent to another location or returned back to biorepository.",
    )
    comments = Column(String, doc="Comments on sample testing.")
    diagnosis_verification = Column(
        Enum(
            "Local review not consistent with diagnostic pathology report",
            "Local review consistent with diagnostic pathology report",
            "Not Available",
            "Not Reported",
            "Other",
            name="diagnosis_verification_enum",
        ),
        doc="Indicates whether the local pathology review was consistent with the diagnostic pathology report.",
    )
    intended_assay = Column(
        AssaysEnum,
        doc="The assay that this sample is expected to be used as input for.",
    )

    aliquots = relationship(
        "Aliquot",
        back_populates="sample",
        sync_backref=False,
        viewonly=True,
        doc="Pertaining to a portion (volume or weight) of the whole.",
    )
    collection_event = relationship(
        CollectionEvent, back_populates="samples", sync_backref=False, viewonly=True
    )
    participant = relationship(
        Participant, back_populates="samples", sync_backref=False, viewonly=True
    )
    shipment = relationship(
        Shipment, back_populates="samples", sync_backref=False, viewonly=True
    )


class Aliquot(MetadataModel):
    __tablename__ = "aliquots"

    sample_id = Column(String, ForeignKey(Sample.cimac_id), primary_key=True)
    slide_number = Column(
        String,  # should be Integer
        CheckConstraint(
            "slide_number ~ '^[0-9]{1,2}$'"
        ),  # should be "slide_number >= 0 and slide_number < 100"
        primary_key=True,  # both True allows for use as multi Foreign Key
        doc="Two digit number that indicates the sequential order of slide cuts, assigned by the CIMAC-CIDC Network.",
    )
    quantity = Column(Integer, doc="Quantity of each aliquot shipped.")
    aliquot_replacement = Column(
        Replacement,
        doc="Status of aliquot if replacement is/was requested.",
    )
    aliquot_status = Column(
        Enum(
            "Aliquot Returned",
            "Aliquot Exhausted",
            "Remainder used for other Assay",
            "Aliquot Leftover",
            "Other",
            name="aliquot_status_enum",
        ),
        doc="Status of aliquot used for other assay, exhausted, destroyed, or returned.",
    )
    material_extracted = Column(
        Enum("DNA", "RNA", "cfDNA", "Other", name="material_extracted_enum"),
        doc="The type of biological material that was extracted from this aliquot.",
    )
    extracted_concentration = Column(
        String,
        doc="The concentration of biological material that was extracted from this aliquot.",
    )
    aliquot_amount = Column(
        String, doc="The amount of extracted aliquot shipped. Example: 400 ng"
    )  # RECHECK Should this be moved to Sample or renamed?
    lymphocyte_influx = Column(
        String,
        doc="Extent of lymphocytic infiltration into the tumor stroma or surrounding environment. Example: 2",
    )

    sample = relationship(
        Sample, back_populates="aliquots", sync_backref=False, viewonly=True
    )



__all__ = [
    "ArtifactCreator",
    "ArtifactCategory",
    "BamFile",
    "Bam_baiFile",
    "BinaryFile",
    "CsvFile",
    "ElisaFile",
    "FastaFile",
    "Fastq_gzFile",
    "FcsFile",
    "File",
    "GzFile",
    "ImageFile",
    "JsonFile",
    "MafFile",
    "NGSAssayFiles",
    "NGSUpload",
    "NpxFile",
    "RccFile",
    "TextFile",
    "TsvFile",
    "Upload",
    "UploadStatus",
    "Vcf_gzFile",
    "VcfFile",
    "XlsxFile",
    "YamlFile",
    "ZipFile",
]

from flask import current_app
from sqlalchemy import (
    Boolean,
    CheckConstraint,
    Column,
    Enum,
    ForeignKey,
    ForeignKeyConstraint,
    Index,
    Integer,
    String,
    text,
    UniqueConstraint,
)
from sqlalchemy.dialects.postgresql import JSONB, UUID
from sqlalchemy.orm import relationship
from sqlalchemy.sql.expression import case

from ..models import Users, with_default_session

from .model_core import MetadataModel
from .trial_metadata import ClinicalTrial, Sample, Shipment


ArtifactCreator = Enum(
    "DFCI", "Mount Sinai", "Stanford", "MD Anderson", name="artifact_creator_enum"
)
ArtifactCategory = Enum(
    "Assay Artifact from CIMAC",
    "Pipeline Artifact",
    "Manifest File",
    name="artifact_category_enum",
)
AssayCreator = Enum(
    "DFCI", "Mount Sinai", "Stanford", "MD Anderson", "TWIST", name="assay_creator_enum"
)
UploadStatus = Enum(
    "started",
    "upload-completed",
    "upload-failed",
    "merge-completed",
    "merge-failed",
    name="upload_status_enum",
)


class Upload(MetadataModel):
    __tablename__ = "uploads"
    id = Column(
        Integer,
        autoincrement=True,
        primary_key=True,
        doc="A unique ID to identify this upload.",
    )
    trial_id = Column(
        String,
        ForeignKey(ClinicalTrial.protocol_identifier),
        primary_key=True,  # both True allows for use as multi Foreign Key
    )

    status = Column(
        UploadStatus,
        nullable=False,
        server_default=UploadStatus.enums[0],  # 'started'
        doc="The current status of the upload.",
    )
    token = Column(
        UUID,
        server_default=text("gen_random_uuid()"),
        nullable=False,
        doc="A long, random identifier for this upload.",
    )
    status_details = Column(
        String,
        nullable=True,
        doc="Text containing feedback on why the upload status is what it is.",
    )
    multifile = Column(
        Boolean, nullable=False, doc="Whether the upload contains multiple files."
    )
    gcs_file_map = Column(
        JSONB,
        nullable=True,
        doc="If multifile, object names for the files to be uploaded mapped to upload_placeholder uuids; else null.",
    )
    gcs_xlsx_uri = Column(
        String,
        nullable=True,
        doc="Track the GCS URI of the .xlsx file used for this upload.",
    )
    upload_type = Column(
        String,
        nullable=False,
        doc="The type of upload (pbmc, wes, olink, wes_analysis, ...)",
    )
    assay_creator = Column(
        AssayCreator, nullable=False, doc="Which CIMAC site created the data"
    )
    uploader_email = Column(
        String,
        ForeignKey(Users.email),
        nullable=False,
        doc="Link to the user who created this upload.",
    )
    shipment_manifest_id = Column(
        String,
        nullable=True,
        doc="Only for manifest uploads",
    )

    # Create a GIN index on the GCS object names
    _gcs_objects_idx = Index(
        "upload_gcs_file_map_idx", gcs_file_map, postgresql_using="gin"
    )
    CheckConstraint("multifile or (gcs_file_map is not null)")

    __mapper_args__ = {
        "polymorphic_on": case(
            [(upload_type.in_(["hande", "wes_uploads"]), upload_type)], else_="base"
        ),
        "polymorphic_identity": "base",
    }
    __table_args__ = (
        ForeignKeyConstraint(
            [trial_id, shipment_manifest_id], [Shipment.trial_id, Shipment.manifest_id]
        ),
    )

    def __init__(self, **kwargs):
        if "uploader_email" not in kwargs:
            from ...shared.auth import get_current_user

            kwargs["uploader_email"] = get_current_user().email

        super().__init__(**kwargs)


class NGSUpload(Upload):
    __tablename__ = "ngs_uploads"

    id = Column(
        Integer,
        autoincrement=True,
        primary_key=True,
        doc="A unique ID to identify this upload.",
    )
    trial_id = Column(
        String,
        primary_key=True,  # both True allows for use as multi Foreign Key
    )
    sequencer_platform = Column(
        Enum(
            "Illumina - HiSeq 2500",
            "Illumina - HiSeq 3000",
            "Illumina - NextSeq 550",
            "Illumina - HiSeq 4000",
            "Illumina - NovaSeq 6000",
            "MiSeq",
            name="sequencer_platform_enum",
        ),
        doc="Sequencer Model, e.g. HiSeq 2500, NextSeq, NovaSeq.",
    )
    library_kit = Column(
        Enum(
            "Hyper Prep ICE Exome Express: 1.0",
            "KAPA HyperPrep",
            "IDT duplex UMI adapters",
            "TWIST",
            name="library_kit_enum",
        ),
        doc="The library construction kit.",
    )
    paired_end_reads = Column(
        Enum("Paired", "Single", name="paired_end_reads_enum"),
        doc="Indicates if the sequencing was performed paired or single ended.",
    )

    __table_args__ = (
        ForeignKeyConstraint([id, trial_id], [Upload.id, Upload.trial_id]),
    )

    records = relationship(
        "NGSAssayFiles", back_populates="upload", sync_backref=False, viewonly=True
    )

    __mapper_args__ = {"polymorphic_identity": "ngs_base"}


class NGSAssayFiles(MetadataModel):
    __tablename__ = "ngs_assay_file_collections"
    id = Column(
        Integer,
        autoincrement=True,
        primary_key=True,
        doc="A unique ID to identify this upload.",
    )

    upload_id = Column(Integer, nullable=False)
    cimac_id = Column(String, nullable=False)
    trial_id = Column(String, nullable=False)

    r1_object_url = Column(
        String,
        doc="Fastq file for the first fragment.",
    )
    r2_object_url = Column(
        String,
        doc="Fastq file for the second fragment.",
    )
    lane = Column(Integer, doc="The lane number from which the reads were generated.")

    bam_object_url = Column(
        String,
        doc="Bam file",
    )
    number = Column(
        Integer,
        doc="An arbitrary number assigned to identify different otherwise equivalent replicates.",
    )

    upload = relationship(
        NGSUpload, back_populates="records", sync_backref=False, viewonly=True
    )
    __table_args__ = (
        ForeignKeyConstraint([trial_id, upload_id], [NGSUpload.trial_id, NGSUpload.id]),
        ForeignKeyConstraint([trial_id, cimac_id], [Sample.trial_id, Sample.cimac_id]),
        CheckConstraint(
            "(r1_object_url is not null and r2_object_url is not null) or bam_object_url is not null"
        ),
        ForeignKeyConstraint(
            [trial_id, upload_id, r1_object_url],
            ["files.trial_id", "files.upload_id", "files.object_url"],
        ),
        ForeignKeyConstraint(
            [trial_id, upload_id, r2_object_url],
            ["files.trial_id", "files.upload_id", "files.object_url"],
        ),
        ForeignKeyConstraint(
            [trial_id, upload_id, bam_object_url],
            ["files.trial_id", "files.upload_id", "files.object_url"],
        ),
    )

    @classmethod
    @with_default_session
    def get_by_id(cls, upload_id, cimac_id, trial_id, session):
        """While a primary key needs to be unique, we want there to be multiple records returned given the ids."""
        with current_app.app_context():
            ret = session.query(cls).filter(
                cls.upload_id == upload_id,
                cls.cimac_id == cimac_id,
                cls.trial_id == trial_id,
            )
        return ret

    @property
    def r1(self):
        return Fastq_gzFile.get_by_id(self.r1_object_url)

    @property
    def r2(self):
        return Fastq_gzFile.get_by_id(self.r2_object_url)

    @property
    def bam(self):
        return BamFile.get_by_id(self.bam_object_url)


class File(MetadataModel):
    __tablename__ = "files"
    object_url = Column(String, primary_key=True)
    upload_id = Column(Integer, primary_key=True)
    trial_id = Column(String, primary_key=False)
    local_path = Column(String, unique=True, doc="Path to a file on a user's computer.")

    upload_placeholder = Column(
        String, doc="A placeholder for when artifact file is being uploaded."
    )
    artifact_creator = Column(
        ArtifactCreator, doc="The name of the center that created this artifact."
    )
    uploader = Column(String, doc="The name of the person uploading the artifact.")
    file_name = Column(
        String,
        doc="The name of the file with extension. Generated from 'gcs_uri_format' fields in templates.",
    )
    uploaded_timestamp = Column(
        String, doc="Timestamp of when artifact was loaded into the system."
    )
    file_size_bytes: Column(Integer, doc="File size in bytes.")
    md5_hash = Column(
        String, doc="MD5 Hash of artifact. Not available for composite GCS objects."
    )
    crc32_hash = Column(String, doc="CRC32c Hash of artifact.")
    visible: Column(
        Boolean,
        doc="Indicates if the artifact is visible. If set to false, the artifact is effectively deleted.",
    )
    artifact_category = Column(ArtifactCategory, doc="Artifact category.")
    data_format = Column(String, doc="Data Format.")
    facet_group = Column(String, doc="The internal data category for this artifact")

    __table_args__ = (
        ForeignKeyConstraint(
            [trial_id, upload_id],
            [Upload.trial_id, Upload.id],
        ),
        UniqueConstraint(trial_id, upload_id, object_url),
    )
    __mapper_args__ = {"polymorphic_on": data_format, "polymorphic_identity": "base"}


# As the subclasses below do NOT have __tablename__ defined
# they will all be included in `files` with any subclass-specific
# columns being NULL for any other object.

# To use a separate table that will be JOINed, define a table name and add
# object_url = Column(String, ForeignKey(<parent>.object_url), primary_key=True)


class FastaFile(File):
    __mapper_args__ = {"polymorphic_identity": "fasta"}


class Fastq_gzFile(File):
    __mapper_args__ = {"polymorphic_identity": "fastq.gz"}


class Vcf_gzFile(File):
    __mapper_args__ = {"polymorphic_identity": "vcf.gz"}


class ImageFile(File):
    __mapper_args__ = {"polymorphic_identity": "image"}


class VcfFile(File):
    __mapper_args__ = {"polymorphic_identity": "vcf"}


class CsvFile(File):
    __mapper_args__ = {"polymorphic_identity": "csv"}


class TsvFile(File):
    __mapper_args__ = {"polymorphic_identity": "tsv"}


class XlsxFile(File):
    __mapper_args__ = {"polymorphic_identity": "xlsx"}


class NpxFile(File):
    __mapper_args__ = {"polymorphic_identity": "npx"}


class ElisaFile(File):
    __mapper_args__ = {"polymorphic_identity": "elisa"}


class BamFile(File):
    __mapper_args__ = {"polymorphic_identity": "bam"}


class Bam_baiFile(File):
    __mapper_args__ = {"polymorphic_identity": "bam.bai"}


class MafFile(File):
    __mapper_args__ = {"polymorphic_identity": "maf"}


class BinaryFile(File):
    __mapper_args__ = {"polymorphic_identity": "binary"}


class TextFile(File):
    __mapper_args__ = {"polymorphic_identity": "text"}


class ZipFile(File):
    __mapper_args__ = {"polymorphic_identity": "zip"}


class FcsFile(File):
    __mapper_args__ = {"polymorphic_identity": "fcs"}


class GzFile(File):
    __mapper_args__ = {"polymorphic_identity": "gz"}


class RccFile(File):
    __mapper_args__ = {"polymorphic_identity": "rcc"}


class JsonFile(File):
    __mapper_args__ = {"polymorphic_identity": "json"}


class YamlFile(File):
    __mapper_args__ = {"polymorphic_identity": "yaml"}



__all__ = [
    "HandeImage",
    "HandeRecord",
    "HandeUpload",
    "WESRecord",
    "WESUpload",
]

from sqlalchemy import (
    CheckConstraint,
    Column,
    Date,
    Enum,
    ForeignKeyConstraint,
    Integer,
    Numeric,
    String,
)
from sqlalchemy.orm import relationship

from .file_metadata import ImageFile, NGSAssayFiles, NGSUpload, Upload
from .model_core import MetadataModel
from .trial_metadata import Sample


class HandeUpload(Upload):
    records = relationship(
        "HandeRecord", back_populates="upload", sync_backref=False, viewonly=True
    )
    images = relationship(
        "HandeImage", back_populates="upload", sync_backref=False, viewonly=True
    )

    __mapper_args__ = {"polymorphic_identity": "hande"}


class HandeImage(ImageFile):
    __tablename__ = "hande_images"
    __mapper_args__ = {"polymorphic_identity": "hande_image.svs"}

    object_url = Column(String, primary_key=True)
    upload_id = Column(Integer, nullable=False)

    upload = relationship(
        HandeUpload, back_populates="images", sync_backref=False, viewonly=True
    )

    __table_args__ = (
        ForeignKeyConstraint(
            [upload_id, object_url], [ImageFile.upload_id, ImageFile.object_url]
        ),
    )


class HandeRecord(MetadataModel):
    __tablename__ = "hande_records"
    upload_id = Column(Integer, nullable=False, primary_key=True)
    cimac_id = Column(String, nullable=False, primary_key=True)
    trial_id = Column(String, nullable=False)
    image_url = Column(String, nullable=False)

    tumor_tissue_percentage = Column(
        Numeric,
        CheckConstraint(
            "tumor_tissue_percentage >= 0 and tumor_tissue_percentage <= 100"
        ),
    )
    viable_tumor_percentage = Column(
        Numeric,
        CheckConstraint(
            "viable_tumor_percentage >= 0 and viable_tumor_percentage <= 100"
        ),
    )
    viable_stroma_percentage = Column(
        Numeric,
        CheckConstraint(
            "viable_stroma_percentage >= 0 and viable_stroma_percentage <= 100"
        ),
    )
    necrosis_percentage = Column(
        Numeric,
        CheckConstraint("necrosis_percentage >= 0 and necrosis_percentage <= 100"),
    )
    fibrosis_percentage = Column(
        Numeric,
        CheckConstraint("fibrosis_percentage >= 0 and fibrosis_percentage <= 100"),
    )
    comment = Column(String)

    upload = relationship(
        HandeUpload, back_populates="records", sync_backref=False, viewonly=True
    )

    __table_args__ = (
        ForeignKeyConstraint(
            [trial_id, upload_id], [HandeUpload.trial_id, HandeUpload.id]
        ),
        ForeignKeyConstraint([trial_id, cimac_id], [Sample.trial_id, Sample.cimac_id]),
    )

    @property
    def image(self) -> HandeImage:
        return HandeImage.get_by_id(self.image_url, self.upload_id)


class WESUpload(NGSUpload):
    __tablename__ = "wes_uploads"

    id = Column(
        Integer,
        autoincrement=True,
        primary_key=True,
        doc="A unique ID to identify this upload.",
    )
    trial_id = Column(
        String,
        primary_key=True,  # both True allows for use as multi Foreign Key
    )
    sequencing_protocol = Column(
        Enum(
            "Express Somatic Human WES (Deep Coverage) v1.1",
            "Somatic Human WES v6",
            "TWIST",
            name="sequencing_protocol_enum",
        ),
        doc="Protocol and version used for the sequencing.",
    )
    bait_set = Column(
        Enum(
            "whole_exome_illumina_coding_v1",
            "broad_custom_exome_v1",
            "TWIST Dana Farber Custom Panel",
            "TWIST Custom Panel PN 101042",
            name="bait_set_enum",
        ),
        nullable=False,
        doc="Bait set ID.",
    )
    read_length = Column(
        Integer,
        CheckConstraint("read_length > 0 and read_length <= 1000"),
        nullable=False,
        doc="Number of cycles for each sequencing read.",
    )

    __table_args__ = (
        ForeignKeyConstraint([id, trial_id], [NGSUpload.id, NGSUpload.trial_id]),
    )

    records = relationship(
        "WESRecord", back_populates="upload", sync_backref=False, viewonly=True
    )

    __mapper_args__ = {"polymorphic_identity": "wes"}


class WESRecord(MetadataModel):
    __tablename__ = "wes_records"

    upload_id = Column(Integer, nullable=False, primary_key=True)
    cimac_id = Column(String, nullable=False, primary_key=True)
    trial_id = Column(String, nullable=False)

    sequencing_date = Column(Date, doc="Date of sequencing.")
    quality_flag = Column(
        Numeric,
        doc="Flag used for quality.",
    )

    upload = relationship(
        WESUpload, back_populates="records", sync_backref=False, viewonly=True
    )

    __table_args__ = (
        ForeignKeyConstraint(
            [trial_id, upload_id], [HandeUpload.trial_id, HandeUpload.id]
        ),
        ForeignKeyConstraint([trial_id, cimac_id], [Sample.trial_id, Sample.cimac_id]),
    )

    @property
    def files(self) -> NGSAssayFiles:
        return NGSAssayFiles.get_by_id(
            upload_id=self.upload_id, cimac_id=self.cimac_id, trial_id=self.trial_id
        )
